/*
	Copyright (C) 2005 by Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the license.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Henner Zilibowitz on 1/01/05.
 *
 *  A test bed for a game. Can support multiple views and players.
 */

#ifndef CARGAME_H
#define CARGAME_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "Player.h"
#include "ScView.h"
#include "Tileset.h"
#include "ScParticles.h"
#include "NeedleImage.h"
#include "StaticImage.h"
#include "PlayerBomb.h"
#include "Minimap.h"
#include "StartLight.h"
#include <list>
#include <vector>
using namespace std;

// block
// Used in algorithm for improving collision detection with tiles by simplifying
// local tile geometry.

struct block
{
    block(int inRight, int inLeft, int inTop, int inBottom)
        : right(inRight), left(inLeft), top(inTop), bottom(inBottom) {}
    
    int right, left, top, bottom;
};

// ViewFollow
// A view and the player it is following

struct ViewFollow
{
	ViewFollow(ScView* inView, short inFollow) : view(inView), follow(inFollow) {}
	ScView* view;
	short follow;
};

enum CameraMode {
   kRotoZoomingCamera,
   kZoomingCamera,
   kFixedCamera,
   kNumCameraModes
};

// forwards declarations
class Wheel;
class WheelSprite;

// class CarGame

class CarGame
{
friend class PathLine;
public:
            CarGame(Tileset* inTilemap, Player** inPlayers, int inNumPlayers, FMOD::System* FMOD_system, Tuple inStartLineA, Tuple inStartLineB, bool inClockwise, const list<Tuple> *inPathPolygon, const list<RoadSignInterval> *inSignIntervals);
   virtual ~CarGame();
	
	virtual void	addView(ScPortRect& portRect, int playerFollow);
	void	doEvent(PlayerEvent* event, unsigned short playerNum = 0);
	virtual void	step(double dt);
	virtual void	renderEveryTexture();
   
	Tileset& getTilemap() const { return tilemap; }
	list<Player*>& getPlayersInGame() { return playersInGame; }
	list<ViewFollow>& getViews() { return views; }
	
   virtual bool   isPlayerHuman(int i);
   
   void changeCameraMode() { cameraMode = (CameraMode)(((int)cameraMode + 1)%(int)kNumCameraModes); }
   
protected:
	Tileset&	tilemap;
	ScAnimation animation;
	Player**	players;
	const int	numPlayers;
	ScRect		gameRect;
	list<ViewFollow>	views;
	list<Player*>	playersInGame;
	ScParticles*	particles;
	NeedleImage*	needle;
	StaticImage*	labelKPH;
	StaticImage*	labelRPM;
	StartLight*	lights[kNumStartLights];
	double		runningTime;
	Minimap*	minimap;
   ScSprite *signsSprite;
	Tuple		startLineA, startLineB;
   bool  clockwise;
	int			lastLight;
	const list<Tuple> *pathPolygon;
	const list<RoadSignInterval> *signIntervals;
	FMOD::Sound *engineSound, *crashSound, *dustSlideSound, *organC6Sound, *organC7Sound;
	FMOD::System* FMOD_System;
	double totalPathLength;
   CameraMode cameraMode;
   
   virtual void    collisionAction(list<Contact> &contacts, Player* playerI, Player* playerJ) {}
	void	computePhysics(double dt);
	void	drawLights();
	virtual void	drawViews(double dt);
	virtual void	drawTime(double theTime);
	void	findTileContacts(ToyCar& car, list<Contact>& contacts);
	void	sewSeams(list<block>& blocks);
	void	resolveCollisions(list<Contact>& contacts);
	void	generateDustFromWheel(const Wheel& wheel, double dt, double rewind);
	virtual void drawDisplay(Player *player, const ScRect& viewRect, double dt);
	void	drawSpeedometer(Player *player);
	void	drawRPMMeter(Player *player);
	void	drawLapCount(Player *player);
	void	drawFinishTime(Player *player);
   void  drawSign(Player *player, const ScRect& viewRect, double dt);
	void	setupStart();
   double computeNearestPathPosition(Tuple pos);
   const RoadSignInterval *findRoadSignInterval(Player *player);
   double computeTotalPathLength();
   double computeCameraRotation(const ScView *view, const Player *player, double dt);
   double computeCameraZoom(const ScView *view, const Player *player, double dt);
};

#endif
