/*
	Copyright (C) 2005 by Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the license.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/
/*
 *  Created by Ruben Henner Zilibowitz on 17/12/05.
 */

#include "HotPotatoGame.h"
#include "GameConstants.h"
#include "glf.h"
#include "TcPreferences.h"

/* When using MINGW (-mno-cygwin), srandom/random isn't defined */
#ifndef seed48
unsigned short * seed48(unsigned short xseed[3]);
#endif
#ifndef drand48
double drand48();
#endif

const double kExplosionSoundFrequency = 44.1e3;	// Hz

// external functions
void ERRCHECK(FMOD_RESULT result);

extern TcPreferences *gPreferences;

HotPotatoGame::HotPotatoGame(Tileset* inTilemap, Player** inPlayers, int inNumPlayers, FMOD::System* inFMOD_system, Tuple inStartLineA, Tuple inStartLineB, bool inClockwise, double inFuseTime, const list<Tuple> *inPathPolygon, const list<RoadSignInterval> *inSignIntervals)
    :   CarGame(inTilemap, inPlayers, inNumPlayers, inFMOD_system, inStartLineA, inStartLineB, inClockwise, inPathPolygon, inSignIntervals),
        bombFuseLength(inFuseTime)
{
    // load and create bomb
    playerWithBomb = players[0];
    bomb = new PlayerBomb(bombFuseLength, playerWithBomb, this, &animation);
    animation.push_back(bomb);
	
	// load sound
#if HAVE_FMOD_LIB
	if (gPreferences->getSounds()) {
		FMOD_RESULT      result;
		result = FMOD_System->createSound("sounds/Explosion.ogg", FMOD_SOFTWARE | FMOD_3D, 0, &explosionSound);
		ERRCHECK(result);
		result = crashSound->set3DMinMaxDistance(4.0f, 10000.0f);
		ERRCHECK(result);
	}
#endif
}

HotPotatoGame::~HotPotatoGame()
{
#if HAVE_FMOD_LIB
	if (gPreferences->getSounds()) {
		FMOD_RESULT      result;
		/*
			Release sounds
		*/
		result = explosionSound->release();
		ERRCHECK(result);
	}
#endif
}

void HotPotatoGame::renderEveryTexture()
{
    int i;
    CarGame::renderEveryTexture();
        // render all the bomb sprite frames
    bomb->moveTo(0,0);
    for (i = 0; i < bomb->getNumFrames(); i++)
    {
        bomb->setFrame(i);
        bomb->doDraw();
    }
}

void HotPotatoGame::collisionAction(list<Contact> &contacts, Player* playerI, Player* playerJ)
{
    // check if the collision is hard enough to switch the bomb between cars
    if (-contacts.back().compute_velocity() > kMinTagVelocity)
    {
        if (playerWithBomb == playerI)
            playerWithBomb = playerJ;
        else if (playerWithBomb == playerJ)
            playerWithBomb = playerI;
        bomb->follow(playerWithBomb);
    }
}

void HotPotatoGame::drawTime(double theTime)
{
    CarGame::drawTime(bombFuseLength - theTime);
}

void HotPotatoGame::detonate(Player* player)
{
	Tuple vec;
	
	playersInGame.remove(player);
	player->car.setVisible(false);
	
	vec = player->car.getLocation_wc() * kScale;
	for (int i = 0; i < 2200; i++)
		particles->newParticle(Particle(vec.x, vec.y, 8.0, (drand48()-0.5)*20.0, (drand48()-0.5)*20.0, kScale));
	
	for (list<Player*>::iterator i = playersInGame.begin(); i != playersInGame.end(); i++)
	{
		// reciprocal of distance
		vec = (*i)->car.getLocation_wc() - player->car.getLocation_wc();
		vec = (40.0/((vec*vec))) * vec;
		(*i)->car.applyImpulse(vec);
	}
	
#if HAVE_FMOD_LIB
	if (gPreferences->getSounds()) {
		// stop the engine sound
		player->setPausedSounds(true);
		
		// play the explosion sound
		FMOD_VECTOR pos = {  player->car.getLocation_wc().x,  player->car.getLocation_wc().y,  0.0f };
		FMOD_VECTOR vel = {  0.0f,  0.0f,  0.0f };
		FMOD::Channel	*explosionChannel;
		FMOD_RESULT result;
		
		result = FMOD_System->playSound(FMOD_CHANNEL_FREE, explosionSound, false, &explosionChannel);
		ERRCHECK(result);
		result = explosionChannel->set3DAttributes(&pos, &vel);
		ERRCHECK(result);
		result = explosionChannel->setVolume(1.0);
		ERRCHECK(result);
		result = explosionChannel->setFrequency(kExplosionSoundFrequency);
		ERRCHECK(result);
	}
#endif
}

void HotPotatoGame::drawDisplay(Player *player, const ScRect& viewRect, double dt)
{
	drawSpeedometer(player);
	drawRPMMeter(player);
	if (runningTime > bombFuseLength) {
		if (playerWithBomb == player) {
			drawLose(player);
		}
		else {
			drawWin(player);
		}
	}
	drawLights();
   drawSign(player, viewRect, dt);
}

void HotPotatoGame::drawLose(Player *player)
{
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glColor4f(1.0f,1.0f,1.0f,0.5f);
	glPushMatrix();
	glScalef(12, 12, 1);
	glTranslatef(2, gPreferences->getScreenHeight()/24.0 + 1, 0);
	glfDrawSolidString("You lose");
	glPopMatrix();
}

void HotPotatoGame::drawWin(Player *player)
{
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glColor4f(1.0f,1.0f,1.0f,0.5f);
	glPushMatrix();
	glScalef(12, 12, 1);
	glTranslatef(2, gPreferences->getScreenHeight()/24.0 + 1, 0);
	glfDrawSolidString("You win");
	glPopMatrix();
}
