/*
	Copyright (C) 2005 by Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the license.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Henner Zilibowitz on 2/01/05.
 */

#include "Player.h"
#include "CarGame.h"
#include <cstdio>

#ifndef INFINITY
   #define INFINITY HUGE_VAL
#endif

// constants

const double kPaddleSteeringSensitity = 0.001;
const double kPaddleThrottleSensitity = 0.001;
const double kPaddleBrakeSensitity = 0.001;

const double kSteerSpeed = kMaxSteerAngle*1.6;		// rad / s
const double kSteerReturnSpeed = kSteerSpeed*10;	// rad / s

const double kThrottleSpeed = 2.0;						// units / s
const double kThrottleReturnSpeed = kThrottleSpeed*10;	// units / s

const double kBrakeSpeed = 4.0;						// units / s
const double kBrakeReturnSpeed = kBrakeSpeed*10;	// units / s

const double kEngineSoundPlayFrequency = 44.1e3;	// Hz
const double kDustSoundPlayFrequency = 22.05e3;	// Hz

// external functions
void ERRCHECK(FMOD_RESULT result);

// local functions

// computes the circuit length defined by a list of points
double pathLength(const list<Tuple> *path)
{
	list<Tuple>::const_iterator i, j;
	double dist = 0;
	for (i = path->begin(), ++(j = path->begin()); i != path->end(); i++, j++) {
		if (j == path->end())
			j = path->begin();
		dist += (*j - *i).Length();
	}
	return dist;
}

// member functions

Player::Player(ToyCar& inCar, list<Tuple> *path, int inPlayerNumber, FMOD::System* inFMOD_System)
	: car(inCar), FMOD_System(inFMOD_System), lapFlag(0), laps(-1), curLap(-1e-1), prevLap(-1e-1), revolutions(-1), bestLapTime(INFINITY), courseTime(0.0),
	  pathEnd(path->end()), pathBegin(path->begin()), pathClosestPerp(INFINITY), pathSize(path->size()), actualPathSize(pathLength(path)), playerNumber(inPlayerNumber),
     lastRoadSignInterval(NULL), signFade(1), nearestPathPos(0), rightWay(true)
{
	// init key states
	for (int i = 0; i < kNumKeys; i++)
		keys[i] = false;
   
   playerType = kPlayerType;
}

double Player::getActualPathDistance() const
{
	int reducedIndex = pathIndex % pathSize;
	list<Tuple>::const_iterator curP = pathEnd;
	Tuple temp;
	curP--;	// moves curP to last point in the circuit
	double dist = 0;
	
	while (reducedIndex < 0)
		reducedIndex += pathSize;
	
	// sums up the portion of the path up to player's closest path segment
	for (int i = 0; i < reducedIndex; i++) {
		temp = *curP;
		curP++;
		if (curP == pathEnd)
			curP = pathBegin;
		dist += (*curP - temp).Length();
	}
	
	// add in the final bit to the players closest point on the path
	dist += (pathClosestPoint - *curP).Length();
	
	// add in amount for the laps
	dist += (pathIndex/pathSize - (pathIndex < 0 ? 1 : 0)) * actualPathSize;
	
	return dist;
}

void Player::computePathLocation()
{
	list<Tuple>::const_iterator Va, Vb;
	int n;
	
	pathClosestPerp = INFINITY;
	for (Va = pathBegin, (Vb = pathEnd)--, n = 0; Va != pathEnd; Vb = Va++, n++)
	{
		if (testPathPoint(Vb))
			pathIndex = n;
		if (testPathSegment(Va, Vb))
			pathIndex = n;
	}
}

Tuple Player::getPathVector()
{
	if (pathA == pathEnd)
	{
		list<Tuple>::const_iterator tempA;
		(tempA = pathB)++;
		if (tempA == pathEnd)
			tempA++;
		return *tempA - *pathB;
	}
	
	return *pathA - *pathB;
}

// Returns the vector of the path at the point pathDisplacement units down from the current location
// pathDiscplacement is expected to be non-negative
Tuple Player::getPathVectorAt(double pathDisplacement)
{
   list<Tuple>::const_iterator tempA = pathA;
   list<Tuple>::const_iterator tempB = pathB;
	if (tempA == pathEnd)
      tempA++;
	if (tempB == pathEnd)
      tempB++;
   
   if (((*tempA) - pathClosestPoint).Norm() > sqr(pathDisplacement))
      return (*tempA) - (*tempB);
   pathDisplacement -= ((*tempA) - pathClosestPoint).Length();
   
   while (pathDisplacement > 0) {
      tempA++;
      tempB++;
      if (tempA == pathEnd)
         tempA++;
      if (tempB == pathEnd)
         tempB++;
      pathDisplacement -= ((*tempA) - (*tempB)).Length();
   }
   
   return (*tempA) - (*tempB);
}

bool Player::testPathPoint(list<Tuple>::const_iterator Vb)
{
	Tuple loc = car.getLocation_wc();
	Tuple perp = loc - *Vb;
	
	if (perp*perp < pathClosestPerp) {
		pathClosestPerp = perp*perp;
		pathClosestPoint = *Vb;
		pathSegmentDistance = 0;
		(pathA = Vb)++;
		if (pathA == pathEnd)
			pathA = pathBegin;
		pathB = Vb;
		return true;
	}
	return false;
}

bool Player::testPathSegment(list<Tuple>::const_iterator Va, list<Tuple>::const_iterator Vb)
{
	Tuple line = *Va - *Vb;
	Tuple loc = car.getLocation_wc();
	double dist = ((loc - *Vb)*line) / (line*line);
	Tuple perp = loc - (dist*line + (*Vb));
	
	if (dist >= 0 && dist < 1 && perp*perp < pathClosestPerp) {
		pathClosestPerp = perp*perp;
		pathClosestPoint = dist*line + (*Vb);
		pathSegmentDistance = dist;
		pathA = Va;
		pathB = Vb;
		return true;
	}
	return false;
}

void Player::updatePathLocation()
{
	list<Tuple>::const_iterator Va, Vb;
	const list<Tuple>::const_iterator curA = pathA;
	const list<Tuple>::const_iterator curB = pathB;
	const int curIndex = pathIndex;
	
	pathClosestPerp = INFINITY;
	
	if (curA == pathEnd)
	{
		testPathPoint(curB);
		
		(Va = curB)++;
		Vb = curB;
		if (Va == pathEnd)	Va = pathBegin;
 		testPathSegment(Va, Vb);
		Va = curB;		
		if (curB == pathBegin)
			(Vb = pathEnd)--;
		else
    		(Vb = curB)--;
		if (testPathSegment(Va, Vb))
			pathIndex = curIndex - 1;
	}
	else
	{
		Va = curA;
		Vb = curB;
		testPathSegment(Va, Vb);
		testPathPoint(Vb);
		if (testPathPoint(Va) == true)
			pathIndex = curIndex + 1;
		
		(Va = curA)++;
		(Vb = curB)++;
		if (Va == pathEnd)	Va = pathBegin;
		if (Vb == pathEnd)	Vb = pathBegin;
		
		if (testPathSegment(Va, Vb))
			pathIndex = curIndex + 1;
		
		if (curA == pathBegin)
			(Va = pathEnd)--;
		else
		    (Va = curA)--;
		
		if (curB == pathBegin)
			(Vb = pathEnd)--;
		else
			(Vb = curB)--;

		if (testPathSegment(Va, Vb))
			pathIndex = curIndex - 1;
	}
}

void Player::doEvent(PlayerEvent* event)
{
	//assert(event != NULL);
	
	switch (event->type)
	{
		case kPlayerKeyUp:
			keyUp(((PlayerEventKey*)event)->key);
			break;
		
		case kPlayerKeyDown:
			keyDown(((PlayerEventKey*)event)->key);
			break;
		
		case kPlayerPaddleMove:
			paddleMove(((PlayerEventPaddle*)event)->control, ((PlayerEventPaddle*)event)->value);
			break;
	}
	
	//printf("player %x received an event\n", this);
}

void Player::paddleMove(PlayerControl which, short value)
{
	switch (which)
	{
		case kSteerWheel:
			car.setSteerWheelAngle(kMaxSteerAngle*(value / 32768.0));
			break;
			
		case kThrottle:
			car.setThrottle(value / 32768.0);
			//printf("throt val: %d, %f\n", value, car.getThrottle());
			break;
			
		case kBrake:
			car.setBrake((value / 32768.0)*0.5 + 0.5);
			break;
		
		default:
			fprintf(stderr, "Player::paddleMove received unknown control type %d\n", which);
	}
/*
	switch (which)
	{
		case kSteerWheel:
			car.setSteerWheelAngle(car.getSteerWheelAngle() + offset*kPaddleSteeringSensitity);
			break;
			
		case kThrottle:
			car.setThrottle(car.getThrottle() + offset*kPaddleThrottleSensitity);
			break;
			
		case kBrake:
			car.setBrake(car.getBrake() + offset*kPaddleBrakeSensitity);
			break;
		
		default:
			fprintf(stderr, "Player::paddleMove received unknown control type %d\n", which);
	}
*/
}

/* Updates the lap counter for this player. Returns 1 or 0 depending on whether
   the call to this function resulted in the lap count being increased or not.
*/
char Player::updateLaps(const Tuple& lineA, const Tuple& lineB, double curTime, bool clockwise)
{
   const Tuple PA = clockwise ? lineB : lineA;
   const Tuple PB = clockwise ? lineA : lineB;
	const Tuple vec1 = car.getLocation_wc()*kScale - PA;
	const Tuple vec2 = PB - PA;
	
	prevLap = curLap;
	curLap = vec2^vec1;
	if (clockwise) curLap = -curLap;
   
	if (vec1*vec2 < 0)
		return 0;
	
	// check if we crossed over the branch cut in the atan function
	if (curLap > 0 && prevLap <= 0) {
        lapFlag++;
        revolutions++;
    }
    else if (curLap < 0 && prevLap >= 0) {
    	lapFlag--;
        revolutions--;
    }
	
	if (lapFlag == 1)
	{
		laps++;			// player has done a lap
		lapFlag = 0;
		
		// update time parameters
		if (laps > 0)
			bestLapTime = std::min((curTime - courseTime), bestLapTime);
		courseTime = curTime;
		//printf("courseTime %lf, bestLapTime %lf\n", courseTime, bestLapTime);
		// test
		//printf("course time: %f\nbest lap time: %f\n", courseTime, bestLapTime);
        
        return 1;
	}
	else if (lapFlag == -2)
	{
		lapFlag = -1;	// player has done a lap backwards
	}
    
    return 0;
}

void Player::dampSteering(double dt)
{
	const double kSteerDampThreshold = 0.2*kMaxSteerAngle;
	double x = car.getSteerWheelAngle();
	if (fabs(x) < kSteerDampThreshold)
		car.setSteerWheelAngle(x*exp(-2*dt));
}

void Player::integrateKeys(double dt)
{
	integrateSteerKeys(dt);
	integrateThrottleKeys(dt);
	integrateBrakeKeys(dt);
}

void Player::integrateSteerKeys(double dt)
{
	// steering
	
	if (keys[kSteerWheelLeft])
	{
		car.setSteerWheelAngle(car.getSteerWheelAngle() + kSteerSpeed * dt);
	}
	
	if (keys[kSteerWheelRight])
	{
		car.setSteerWheelAngle(car.getSteerWheelAngle() - kSteerSpeed * dt);
	}
	
	if (!(keys[kSteerWheelLeft] || keys[kSteerWheelRight]))
	{
		double x = car.getSteerWheelAngle();
		reduce_by_at_most(x, kSteerReturnSpeed*dt);
		car.setSteerWheelAngle(x);
	}
}

void Player::integrateThrottleKeys(double dt)
{
	// throttle
	// with auto-braking
	
	keys[kBrakeMore] = false;
	if (keys[kThrottleMore])
	{
		car.setThrottle(car.getThrottle() + kThrottleSpeed * dt);
		if (car.getVelocity().x < 0)
			keys[kBrakeMore] = true;
	}
	
	if (keys[kThrottleLess])
	{
		car.setThrottle(car.getThrottle() - kThrottleSpeed * dt);
		if (car.getVelocity().x > 0)
			keys[kBrakeMore] = true;
	}
	
	if (!(keys[kThrottleMore] || keys[kThrottleLess]))
	{
		double x = car.getThrottle();
		reduce_by_at_most(x, kThrottleReturnSpeed*dt);
		car.setThrottle(x);
	}
}

void Player::integrateBrakeKeys(double dt)
{
	// brake
	
	if (keys[kBrakeMore])
	{
		car.setBrake(car.getBrake() + kBrakeSpeed * dt);
	}
	
	if (keys[kBrakeLess])
	{
		car.setBrake(car.getBrake() - kBrakeSpeed * dt);
	}
	
	if (!(keys[kBrakeMore] || keys[kBrakeLess]))
	{
		double x = car.getBrake();
		reduce_by_at_most(x, kBrakeReturnSpeed*dt);
		car.setBrake(x);
	}
}

void Player::reset(const Tuple& lineA, const Tuple& lineB, const Player *winner, bool clockwise)
{
   const Tuple PA = clockwise ? lineB : lineA;
   const Tuple PB = clockwise ? lineA : lineB;
	const Tuple vec1 = car.getLocation_wc()*kScale - PA;
	const Tuple vec2 = PB - PA;

   curLap = vec2^vec1;//vec2.Angle(vec1);
   if (clockwise)
      curLap = -curLap;
   prevLap = curLap;

   lapFlag = 0;
   laps = 0;
   revolutions = 0;

   bestLapTime = INFINITY;
   courseTime = 0;

   if (this != winner)
   {
      computePathLocation();
      pathIndex += (winner->pathIndex / pathSize) * pathSize;
      while (winner->pathIndex - pathIndex > pathSize/2)
         pathIndex += pathSize;
      while (pathIndex - winner->pathIndex > pathSize/2)
         pathIndex -= pathSize;
   }
}

const double kStuckTestThrottle = 0.8;
const double kStuckTestStepSize = 0.01;
const double kStuckTestTimePeriod = 0.8;
const double kStuckTestMaxVelocity = 5.0;

// Does a short look ahead calculation to see if the player is
// obstructed in front by wall in the tilemap.
bool Player::playerIsStuck()
{
	const Tuple p = car.getLocation_wc() - car.getNorm().perp()*(car.getGeomRadius() + 0.1);
	return carGame->getTilemap().getTileAtPoint(p.x*kScale, p.y*kScale) == kWallTile;
}

void Player::printDebugInfo() const
{
	//printf("player id: %x\n", (unsigned int)(this));
	printf("player id: %x\n", (long long)(this));
	printf("player number: %x\n", getPlayerNumber());
	printf("stuck: %d\n", ((Player*)(this))->playerIsStuck());
}

double sqr(double x)
{
	return x*x;
}

// This function is used to protect us from unintended channel stealing by FMOD.
// This could happen if we managed to exceed the maximum number of allowed channels
// we initialised with.
bool channelIsValid(FMOD::Channel* channel)
{
#if HAVE_FMOD_LIB
	int index;
	FMOD_RESULT result;
	if (channel == NULL) {
		fprintf(stderr, "channelIsValid: channel == NULL\n");
		return false;
	}
	result = channel->getIndex(&index);
	if (result != FMOD_OK) {
		fprintf(stderr, "channelIsValid: channel not valid\n");
	}
	return result == FMOD_OK;
#else
	return false;
#endif
}

const double kDustVolumeFactor = 1.0 / 60.0;

void Player::updateSounds()
{
#if HAVE_FMOD_LIB
	FMOD_VECTOR pos;
	FMOD_VECTOR vel;
	FMOD_RESULT result;
	FMOD::Channel *dustChannels[4] = { tireChannelFL, tireChannelFR, tireChannelBL, tireChannelBR };
	const Wheel *wheels[4] = { &car.getWheel_FL(), &car.getWheel_FR(), &car.getWheel_BL(), &car.getWheel_BR() };
	int i;
	Tuple tempVec;
	double tempScalar;
	
	// update engine sound
	
	if (channelIsValid(engineChannel)) {
		pos.x = car.getLocation_wc().x;
		pos.y = car.getLocation_wc().y;
		pos.z = 0;
		
		vel.x = car.getVelocity_wc().x;
		vel.y = car.getVelocity_wc().y;
		vel.z = 0;
		
		result = engineChannel->set3DAttributes(&pos, &vel);
		ERRCHECK(result);
		// n.b. some of the number used below are slightly arbitrary, and probably should be put into constants defined elsewhere.
		// The numbers used seem to produce adaquate sounding engine for now.
		result = engineChannel->setFrequency(0.5 + fabs(car.getRPM()) / 2.5e3 * kEngineSoundPlayFrequency);   // frequency in Hz
		ERRCHECK(result);
		tempScalar = 0.2 + fabs(car.getRPM()) / 3e3;
		if (tempScalar > 1.0)
			tempScalar = 1.0;
		result = engineChannel->setVolume(tempScalar);
		ERRCHECK(result);
	}
	
	// update dust slide sound for each wheel
	
	for (i = 0; i < 4; i++) {
		if (channelIsValid(dustChannels[i])) {
			tempScalar = sqrt(sqr(kLongtitudinalDustFactor*wheels[i]->getLongtitudinalSlipVelocity()) +
							  sqr(kLateralDustFactor*wheels[i]->getVelocity().y)) * kDustVolumeFactor;
			if (tempScalar > 1.0)
				tempScalar = 1.0;
			result = dustChannels[i]->setVolume(tempScalar);
			ERRCHECK(result);
			result = dustChannels[i]->setFrequency((1.0 - 0.4*pow(tempScalar,0.4)) * kDustSoundPlayFrequency);
			ERRCHECK(result);
			tempVec = car.getNorm().fromLocal(wheels[i]->getLocation_vc()) + car.getLocation_wc();
			pos.x = tempVec.x;
			pos.y = tempVec.y;
			tempVec = car.getNorm().fromLocal((car.getVelocity() + (car.getAngularVelocity() * wheels[i]->getLocation_vc().perp())));
			vel.x = tempVec.x;
			vel.y = tempVec.y;
			result = dustChannels[i]->set3DAttributes(&pos, &vel);
			ERRCHECK(result);
		}
	}
#endif
}

const int kEngineChannelPriority = 128;
const int kDustChannelPriority = 128;

void Player::setEngineSound(FMOD::Sound *engineSound)
{
#if HAVE_FMOD_LIB
	FMOD_VECTOR pos = { car.getLocation_wc().x, car.getLocation_wc().y, 0.0f };
	FMOD_VECTOR vel = { car.getVelocity_wc().x, car.getVelocity_wc().y, 0.0f };
	FMOD_RESULT result;
	
	result = FMOD_System->playSound(FMOD_CHANNEL_FREE, engineSound, true, &engineChannel);
	ERRCHECK(result);
	result = engineChannel->set3DAttributes(&pos, &vel);
	ERRCHECK(result);
	result = engineChannel->setPaused(false);
	ERRCHECK(result);
	result = engineChannel->setVolume(0);
	ERRCHECK(result);
	result = engineChannel->setPriority(kEngineChannelPriority);
	ERRCHECK(result);
#endif
}

void Player::setDustSlideSound(FMOD::Sound *dustSlideSound)
{
#if HAVE_FMOD_LIB
	FMOD_VECTOR pos = { 0, 0, 0 };
	FMOD_VECTOR vel = { 0, 0, 0 };
	FMOD_RESULT result;
	FMOD::Channel **dustChannels[4] = { &tireChannelFL, &tireChannelFR, &tireChannelBL, &tireChannelBR };
	int i;
	Tuple temp;
	
	for (i = 0; i < 4; i++) {
		result = FMOD_System->playSound(FMOD_CHANNEL_FREE, dustSlideSound, true, dustChannels[i]);
		ERRCHECK(result);
		result = (*dustChannels[i])->set3DAttributes(&pos, &vel);
		ERRCHECK(result);
		result = (*dustChannels[i])->setPaused(false);
		ERRCHECK(result);
		result = (*dustChannels[i])->setVolume(0);
		ERRCHECK(result);
		result = (*dustChannels[i])->setPriority(kDustChannelPriority);
		ERRCHECK(result);
	}
#endif
}

void Player::setPausedSounds(bool val)
{
#if HAVE_FMOD_LIB
	engineChannel->setPaused(val);
	tireChannelFL->setPaused(val);
	tireChannelFR->setPaused(val);
	tireChannelBL->setPaused(val);
	tireChannelBR->setPaused(val);
#endif
}
