/*
	Copyright (C) 2005 by Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the license.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Henner Zilibowitz on 2/01/05.
 */

#ifndef PLAYER_H
#define PLAYER_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "ToyCar.h"
#include "GameConstants.h"

#if HAVE_FMOD_LIB
#include "fmod.hpp"
#include "fmod_errors.h"
#endif

// PlayerEventType - describes the types of events a player can receive
typedef unsigned char PlayerEventType;

// Constants for different event types
const PlayerEventType kPlayerKeyUp = 0;
const PlayerEventType kPlayerKeyDown = 1;
const PlayerEventType kPlayerPaddleMove = 2;

// PlayerEvent - base class for player events
struct PlayerEvent
{
	PlayerEventType type;
};

// PlayerControl - describes the controls a player has
typedef unsigned char PlayerControl;

// PlayerEventPaddle - an event of a "paddle" changing value. Eg mouse moved or joystick moved.
struct PlayerEventPaddle : public PlayerEvent
{
	PlayerControl control;
	short value;
};

// PlayerKey - describes the keys a player has
typedef unsigned char PlayerKey;

// PlayerEventKey - an event of a key press or release
struct PlayerEventKey : public PlayerEvent
{
	PlayerKey key;
};

// the controls
enum
{
	kSteerWheel = 0,
	kThrottle,
	kBrake
};

// the keys
enum
{
	kSteerWheelLeft = 0,
	kSteerWheelRight,
	kThrottleMore,
	kThrottleLess,
	kBrakeMore,
	kBrakeLess,
	kNumKeys
};

const long kPlayerType = 'PLYR';

// Player class

class CarGame;	// forwards declaration

class Player
{
public:
			Player(ToyCar& inCar, list<Tuple> *path, int inPlayerNumber, FMOD::System* inFMOD_System);
	virtual ~Player() {}
	
	virtual void	doEvent(PlayerEvent* event);
	
	virtual void	dampSteering(double dt);
	virtual void	integrateKeys(double dt);
	virtual void	integrateSteerKeys(double dt);
	virtual void	integrateThrottleKeys(double dt);
	virtual void	integrateBrakeKeys(double dt);
	
	ToyCar&	car;
	
	char	updateLaps(const Tuple& lineA, const Tuple& lineB, double curTime, bool clockwise = false);
	short	getLaps()	{ return laps; }
	double	getBestLapTime()	{ return bestLapTime; }
	double	getCourseTime()		{ return courseTime; }
	double  getCurLap() { return curLap; }
   short	getRevolutions() { return revolutions; }
   void    reset(const Tuple& lineA, const Tuple& lineB, const Player *winner, bool clockwise = false);
	void	updatePathLocation();
	void	computePathLocation();
   
	double getPathDistance() const { return (pathIndex + pathSegmentDistance)/pathSize; }
	Tuple getPathLocation() const { return pathClosestPoint; }
	double getPathPerpDist() const { return pathClosestPerp; }
	Tuple getPathVector();
	Tuple getPathVectorAt(double pathDisplacement);
	int getPlayerNumber() const { return playerNumber; }
	void setPausedSounds(bool val);
	void updateSounds();
	void setEngineSound(FMOD::Sound *engineSound);
	void setDustSlideSound(FMOD::Sound *dustSlideSound);
	void setGame(CarGame *inCarGame) { carGame = inCarGame; }
	
	virtual void printDebugInfo() const;
	
	double getActualPathDistance() const;
	
	int getPathIndex() const { return pathIndex; }
	
   virtual bool isAI() const { return false; }
   
   long getPlayerType() const { return playerType; }
   
   void setLastRoadSignInterval(const RoadSignInterval *s) { lastRoadSignInterval = s; }
   const RoadSignInterval *getLastRoadSignInterval() const { return lastRoadSignInterval; }
   double getSignFade() const { return signFade; }
   void setSignFade(double x) { signFade = x; }
   
   void setNearestPathPos(double x) { nearestPathPos = x; }
   double getNearestPathPos() const { return nearestPathPos; }
   
   void setRightWay(bool x) { rightWay = x; }
   double getRightWay() const { return rightWay; }
   
protected:
	void	keyUp(PlayerKey key) { keys[key] = false; }
	void	keyDown(PlayerKey key) { keys[key] = true; }
	void	paddleMove(PlayerControl which, short value);
	bool	testPathPoint(list<Tuple>::const_iterator Vb);
	bool	testPathSegment(list<Tuple>::const_iterator Va, list<Tuple>::const_iterator Vb);
	bool	playerIsStuck();
	
	FMOD::System* FMOD_System;
	FMOD::Channel *engineChannel, *tireChannelFL, *tireChannelFR, *tireChannelBL, *tireChannelBR;
	CarGame *carGame;
	
   // A unique identifier for the type of player. Subclasses should use a different value for this.
   long playerType;
   
	// key states
	bool	keys[kNumKeys];
	
	// lap counting variables
	short	lapFlag;
	short	laps;
	double	curLap, prevLap;
	short   revolutions;
    
	// times
	double	bestLapTime;
	double	courseTime;
	
	// path information
	const list<Tuple>::const_iterator pathEnd, pathBegin;
	list<Tuple>::const_iterator pathA, pathB;
	int pathIndex;
	double pathSegmentDistance;
	double pathClosestPerp;
	Tuple pathClosestPoint;
	const int pathSize;	// number of nodes in path list
	const double actualPathSize;	// actual length
	const RoadSignInterval *lastRoadSignInterval;
   double signFade;
   double nearestPathPos;
   bool rightWay;
   
	// id number
	const int playerNumber;
};

#endif
