/*
	Copyright (C) 2005 by Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the license.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Henner Zilibowitz on 31/10/07.
 */

#include "RobotHotPotatoPlayer.h"
#include "HotPotatoGame.h"
#include <cassert>
#include <cstdio>

#ifndef INFINITY
   #define INFINITY HUGE_VAL
#endif

RobotHotPotatoPlayer::RobotHotPotatoPlayer(ToyCar& inCar, list<Tuple> *path, int inPlayerNumber, FMOD::System* inFMOD_System)
	: RobotPlayer(inCar, path, inPlayerNumber, inFMOD_System), backupDist(-1.0)
{
   playerType = kRobotHotPotatoPlayerType;
}

const double kMaxPathDistForDirectChase = 5.0;		// m

// this function computes the shortest distance to the given player from this player.
// The return value indicates what kind of route would be the fastest one to take.
// Possible routes would be forwards around the track, backwards around the track, and directly
// to the player.
FastestRoute RobotHotPotatoPlayer::playerDistance(const Player *player, double *dist)
{
	double pathDist = fmod(player->getActualPathDistance() - getActualPathDistance(), actualPathSize);
	FastestRoute route;
	while (pathDist < 0)
		pathDist += actualPathSize;
	if (pathDist > actualPathSize*0.5) {
		pathDist = actualPathSize - pathDist;
		route = kBackwardsTrack;
	}
	else {
		route = kForwardsTrack;
	}
	if (pathDist < kMaxPathDistForDirectChase) {
		*dist = (car.getLocation_wc() - player->car.getLocation_wc()).Length();
		return kDirectLine;
	}
	*dist = pathClosestPerp + pathDist + player->getPathPerpDist();
	return route;
}

void RobotHotPotatoPlayer::integrateKeys(double dt)
{
	const Player * playerWithBomb = dynamic_cast<HotPotatoGame*>(carGame)->getPlayerWithBomb();
	
	if (playerWithBomb != this) {
		// try to avoid the player with the bomb
		avoidPlayer(playerWithBomb, dt);
		
		// set debugging data
		catching = false;
		avoiding = true;
		targetPlayer = playerWithBomb;
	}
	else {
		// find nearest player
		list<Player*> &playersInGame = carGame->getPlayersInGame();
		double minDist = INFINITY, dist;
		FastestRoute fastestRoute = kForwardsTrack, route;
		Player* closestPlayer = NULL;
		for (list<Player*>::iterator i = playersInGame.begin(); i != playersInGame.end(); i++) {
			if (*i == this)
				continue;
			route = playerDistance(*i,&dist);
			if (dist < minDist) {
				minDist = dist;
				fastestRoute = route;
				closestPlayer = *i;
			}
		}
		assert(closestPlayer != NULL);
		
		if (minDist + (car.getVelocity_wc() - closestPlayer->car.getVelocity_wc())*(car.getLocation_wc() - closestPlayer->car.getLocation_wc()).Direction()/kMinTagVelocity < 1.0) {
			// need more speed to catch player here
			// should do something like drive backwards to get a run up perhaps.
			car.setThrottle(0);
			//backup(true, dt);
		}
		else {
			// try to catch them
			catchPlayer(closestPlayer, dt, fastestRoute);
		}
		
		// set debugging data
		catching = true;
		avoiding = false;
		targetPlayer = closestPlayer;
		debugRoute = fastestRoute;
	}
}

const double kHaltingDistance = 3.0;
const double kOrientationWeightingFactor = 3.5;

void RobotHotPotatoPlayer::avoidPlayer(const Player *player, double dt)
{
	// try to avoid the player with the bomb
	//const Tuple seg = *pathA - *pathB;
	//const double orientationWeighting = (seg ^ car.getNorm()) * kOrientationWeightingFactor;
	double pathDist = fmod(player->getActualPathDistance() - getActualPathDistance(), actualPathSize);
	FastestRoute route;
	
	while (pathDist < 0)
		pathDist += actualPathSize;
	if (pathDist /*+ orientationWeighting*/ > actualPathSize*0.5) {
		route = kForwardsTrack;
	}
	else {
		route = kBackwardsTrack;
	}
	
	if (fabs(actualPathSize*0.5 - pathDist) < kHaltingDistance) {
		car.setThrottle(0.0);
		car.setBrake(1.0);
		debugRoute = (FastestRoute)(-1);
	}
	else {
		car.setBrake(0.0);
		assert(route == kForwardsTrack || route == kBackwardsTrack);
		followPath(route == kForwardsTrack, dt);
		debugRoute = route;
	}
}

void RobotHotPotatoPlayer::catchPlayer(const Player *player, double dt, FastestRoute route)
{
	car.setBrake(0.0);
	switch (route) {
		case kForwardsTrack:
			followPath(true, dt);
			break;
		case kBackwardsTrack:
			followPath(false, dt);
			break;
		case kDirectLine:
			driveTowards(player->car.getLocation_wc(), 1.0);
			break;
	}
}

void RobotHotPotatoPlayer::printDebugInfo() const
{
	RobotPlayer::printDebugInfo();
	printf("route: %d\n", debugRoute);
	printf("avoiding: %d\n", avoiding);
	printf("catching: %d\n", catching);
	//printf("targetPlayer: %x\n", (unsigned int)(targetPlayer));
	printf("targetPlayer: %x\n", (long long)(targetPlayer));
	//printf("backupDist: %lf\n", backupDist);
}
