/*
 *  RobotPlayer.cpp
 *  toycars
 *
 *  Created by Ruben Henner Zilibowitz on 5/08/06.
 *  Copyright 2006 Ruben Henner Zilibowitz. All rights reserved.
 *
 */

#include "RobotPlayer.h"
#include <cstdio>

RobotPlayer::RobotPlayer(ToyCar& inCar, list<Tuple> *path, int inPlayerNumber, FMOD::System* inFMOD_System)
	: Player(inCar, path, inPlayerNumber, inFMOD_System), backupDist(0)
{
	// set debug data
	unsticking = false;
   playerType = kRobotPlayerType;
}

void RobotPlayer::integrateKeys(double dt)
{
	followPath(true, dt);
}

void RobotPlayer::driveTowards(Tuple aim, double throttle)
{
	double angle = atan2(aim.y - car.getLocation_wc().y, aim.x - car.getLocation_wc().x);
	
	angle -= car.getOrientation_wc();
	angle = fmod(angle, kTwoPi);
	if (angle > kPi)
		angle -= kTwoPi;
	else if (angle < -kPi)
		angle += kTwoPi;
	Tuple norm = aim - car.getLocation_wc();
	if (fabs(angle) > kPiOn2) {
		car.setThrottle(-throttle);
		car.setSteerWheelAngle(car.getOrientation_wc() + kPi - atan2(norm.y,norm.x));
	}
	else {
		car.setThrottle(throttle);
		car.setSteerWheelAngle(atan2(norm.y,norm.x) - car.getOrientation_wc());
	}
}

// Try to get unstuck by driving towards the path
void RobotPlayer::backup(bool forwards, double dt)
{
	// Try to get unstuck by driving towards the path
	const Tuple aim = 0.5*pathClosestPoint + 0.5*(forwards ? *pathB : *pathA);
	if ((car.getLocation_wc() - aim).Norm() < sqr(1))
		backupDist = -1;
	car.setBrake(0.0);
	driveTowards(aim, 0.75);
	backupDist -= car.getVelocity_wc().Length() * dt;
}

const double kMaxUnstickingBackupDist = 7.0;

void RobotPlayer::followPath(bool forwards, double dt)
{
	if (playerIsStuck()) {
		backupDist = kMaxUnstickingBackupDist;
	}
	if (backupDist > 0) {
		backup(true, dt);
		
		// set debug data
		unsticking = true;
	}
	else if (car.getVelocity().x < -0.5) {
		// stop the reversing
		car.setThrottle(0.0);
		car.setBrake(1.0);
	}
	else {
		// set debug data
		unsticking = false;
		
		car.setBrake(0.0);
		
		// Calculate steering angle
		
		list<Tuple>::const_iterator pathC;
		if (forwards) {
			pathC = pathB;
			pathC++;
			if (pathC == pathEnd)
				pathC = pathBegin;
		}
		else {
			pathC = pathA;
			if (pathC == pathBegin)
				(pathC = pathEnd)--;
			pathC--;
		}
		
		// Steer parallel to path
		Tuple line = *pathC - (forwards ? *pathB : *pathA);
		car.setSteerWheelAngle(atan2(line.y,line.x) - car.getOrientation_wc());
		Tuple norm = pathClosestPoint - car.getLocation_wc();
		
		// Correct to steer towards centre of path
		
		if ((line^norm) < 0)
			car.setSteerWheelAngle(car.getSteerWheelAngle() - norm.Length() * 0.1);
		else
			car.setSteerWheelAngle(car.getSteerWheelAngle() + norm.Length() * 0.1);
		
		//car.mSteerWheelAngle += (line^norm) / line.Length() / kPi * 0.1;
		
		list<Tuple>::const_iterator pathD = pathC;
		if (forwards) {
			pathD++;
			if (pathD == pathEnd)
				pathD = pathBegin;
		}
		else {
			if (pathD == pathBegin)
				(pathD = pathEnd)--;
			else
				pathD--;
		}
		
		// Look ahead and steer around corner slightly in advance
		double pathAngle = line.Angle(*pathD - *pathC);
		double segmentEndDistance = (car.getLocation_wc() - *pathC).Length();
		car.setSteerWheelAngle(car.getSteerWheelAngle() + pathAngle / (segmentEndDistance + 0.5));
		
		if (fabs(pathAngle) > kPi)
			printf("yeah\n");
		
		// Calculate throttle
		
		double temp = 2*fabs(pathAngle) * car.getVelocity().x/(segmentEndDistance + 1.0);
		if (temp > 1.0)
			car.setThrottle(1.0 / temp);
		else
			car.setThrottle(1.0);
	}
   
   double slip = car.getWheel_BL().getLongtitudinalSlipVelocity() + car.getWheel_BR().getLongtitudinalSlipVelocity();
   if (fabs(slip) > 5) {
      car.setThrottle(5 / (-slip));
   }
}

void RobotPlayer::printDebugInfo() const
{
	Player::printDebugInfo();
	printf("unsticking: %d\n", unsticking);
}
