/*
	Copyright (C) 2005 by Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net>
	Part of the Toy Cars Project http://toycars.sourceforge.net

	This program is free software; you can redistribute it and/or modify
	it under the terms of the license.
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY.

	See the COPYING file for more details.
*/

/*
 *  Created by Ruben Henner Zilibowitz on 31/12/04.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <Windows.h>
#include <Mmsystem.h>
#endif

#if defined(__APPLE__) && defined(__MACH__)
#include <OpenGL/gl.h>	// Header File For The OpenGL32 Library
#else

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>	// Header File For The OpenGL32 Library	// Header File For The OpenGL32 Library
#endif

#include "SDL.h"
#include "SDL_image.h"

#include <iostream>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#ifdef WIN32
  #include <direct.h>
  #define chdir _chdir
  #define time(x) 0
#else
  #include <fcntl.h>
  #include <ctime>
#endif
#include <errno.h>
#include <unistd.h>

#include "ScException.h"
#include "CarGame.h"
#include "HotPotatoGame.h"
#include "KnockOutGame.h"
#include "GetSeconds.h"
#include "GameConstants.h"
#include "tinyxml.h"
#include "glf.h"
#include "Menu.h"
#include "TcPreferences.h"
#include "RobotPlayer.h"
#include "RobotHotPotatoPlayer.h"
#include "Controllers.h"

#if HAVE_FMOD_LIB
#include "fmod.hpp"
#include "fmod_errors.h"
#endif

#include "main.h"

#ifndef strtok_r
   #define strtok_r(s,d,p) strtok(s,d)
#endif

//
// Types

struct SearchDirectories {
   char path[256];
   SearchDirectories *next;
};

//
// Constants

const char kLoadingImageFileName[] = "images/title.png";	// path to loading image from within Data dir
const char kTitleSongFileName[] = "music/Motor_Car_Song.ogg";
//const bool kUseVBLSynch = false;	// synchronise with the screen retraces?
const int kFMODMaxChannels = 2048;	// needs to be high enough so that channel stealing does not occur which may crash the game
const int kMaxPlayersInGame = 80;

//
// Globals

char gDataDir[256] = DATA_DIR;		// Path to the Data dir (either absolute or relative to the executable)
char gUserDataDir[256] = DATA_DIR;		// Path to user data and preferences xml file (either absolute or relative to the executable)
SearchDirectories *gSearchPath;
CarGame* gGameP = NULL;
unsigned int gNumPlayers;	// The number of players
unsigned int gNumHumanPlayers;
TcPreferences *gPreferences;

//
// Game Mode enum

enum GameMode
{
	GameMode_TimeTrial,
	GameMode_HotPotato,
	GameMode_KnockEmOut
} gGameMode;

//
// Functions

void freeSearchDirectories(SearchDirectories *dirs) {
   SearchDirectories *temp;
   while (dirs != NULL) {
      temp = dirs->next;
      free(dirs);
      dirs = temp;
   }
}

SearchDirectories *makeSearchDirectories(const char *path, SearchDirectories *next) {
   SearchDirectories *result = new SearchDirectories;
   strncpy(result->path, path, 255);
   result->next = next;
   return result;
}

TiXmlDocument *findXMLFile(const char *filename, SearchDirectories *dirs, char *outPath = NULL)
{
   const int filenameLen = strlen(filename);
   char str[256];
   TiXmlDocument *xmlDoc;
   while (dirs != NULL) {
      if (strlen(dirs->path) + filenameLen + 1 < 255) {
         strcpy(str, dirs->path);
         strcat(str, "/");
         strcat(str, filename);
         xmlDoc = new TiXmlDocument(str);
         if (xmlDoc->LoadFile()) {
            if (outPath)
               strcpy(outPath, dirs->path);
            return xmlDoc;
         }
         else {
            delete xmlDoc;
         }
         dirs = dirs->next;
      }
   }
   return NULL;
}

void ERRCHECK(FMOD_RESULT result)
{
#if HAVE_FMOD_LIB
    if (result != FMOD_OK)
    {
        printf("FMOD error! (%d) %s\n", result, FMOD_ErrorString(result));
        exit(-1);
    }
#endif
}

void PixelDataForSurface(const SDL_Surface* src, ScPixelData& dst)
{
	dst.pitch = src->pitch;
	dst.bytes_pp = src->format->BytesPerPixel;
	dst.width = src->w;
	dst.height = src->h;
	
	switch (dst.bytes_pp)
	{
		case 4:
			dst.format = GL_RGBA;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGBA;
			break;
		
		case 3:
			dst.format = GL_RGB;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGB;
			break;
		
		default:
			ScThrowErr("Only 32 bit or 24 bit is supported at present");
			break;
	}
	
	dst.pixels = (const char*)src->pixels;
}

void DrawSplash()
{
	SDL_Surface* img;
	char string64[64];
	//float minx, miny, maxx, maxy;
    
	// Load the image
	img = IMG_Load(kLoadingImageFileName);
	if (img == NULL) {
		fprintf(stderr, "SDL_image error : %s\n", IMG_GetError());
		ScThrowErr("Could not load splash screen image file");
	}
	
	// Draw the image
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glPixelZoom(1, -1);
	glRasterPos2f(gPreferences->getScreenWidth()*0.5 - img->w*0.5, gPreferences->getScreenHeight()*0.5 + img->h*0.5);
	glClear(GL_COLOR_BUFFER_BIT);
	glDrawPixels(img->w, img->h, GL_RGB, GL_UNSIGNED_BYTE, img->pixels);
	
	// Draw version number
	sprintf(string64, "v. %s", VERSION);
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
	//glfGetStringBounds(string64, &minx, &miny, &maxx, &maxy);
	glTranslatef(gPreferences->getScreenWidth() - 96, 20, 0);
	glScalef(12, 12, 1);
	glfDrawSolidString(string64);
    
	// Free sdl image and swap gl buffers
	SDL_FreeSurface(img);
	//SDL_GL_SwapBuffers();
}

void DrawLoading()
{
    // Draw loading text
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glScalef(24, 24, 1);
	glTranslatef(1, 1, 0);
	glfDrawSolidString("Loading...");
    SDL_GL_SwapBuffers();
}

void DrawPaused()
{
    // Draw loading text
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glScalef(12, 12, 1);
	glTranslatef(1, 1, 0);
	glfDrawSolidString("Paused (press p/start to unpause)...");
    SDL_GL_SwapBuffers();
}

void ShowCredits()
{
    // Draw loading text
	glColor4f(1.0f, 0.0f, 0.0f, 1.0f);
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glTranslatef(10, gPreferences->getScreenHeight()*0.5, 0);
	glScalef(8, 8, 1);
	glfDrawSolidString("Ruben Henner Zilibowitz <rzilibowitz@users.sourceforge.net> - Main Developer");
	glTranslatef(0, -2.5, 0);
	glfDrawSolidString("James Englund with Crash Symphony (http://www.crashsymphony.com/) - The Motor Car Song");
	glTranslatef(0, -2.5, 0);
	glfDrawSolidString("Jeff Hanson - Hairpin track");
	glTranslatef(0, -2.5, 0);
	glfDrawSolidString("FMOD music & sound effects library (http://www.fmod.org/)");
	glTranslatef(0, -2.5, 0);
	glfDrawSolidString("TinyXml library (http://www.grinninglizard.com/tinyxml/)");
	glTranslatef(0, -2.5, 0);
	glfDrawSolidString("GLF library by Roman Podobedov (http://local.wasp.uwa.edu.au/~pbourke/oldstuff/glf/glf.tar.gz)");
    glLoadIdentity();
	glOrtho(0, gPreferences->getScreenWidth(), 0, gPreferences->getScreenHeight(), -1, 1);
	glScalef(10, 10, 1);
	glTranslatef(1, 1, 0);
	glfDrawSolidString("Press any key/button to continue...");
    SDL_GL_SwapBuffers();
	
	SDL_Event event;
	while (1)
	{
		if (SDL_WaitEvent(&event) == 0) {
			printf("error waiting for events\n");
			exit(EXIT_FAILURE);
		}
		if (event.type == SDL_KEYDOWN)
			break;
		if (event.type == SDL_JOYBUTTONDOWN)
			break;
	}
}

/*
* This little groovy formula I cooked up in Mathematica. It returns the shortest line
* distance from the point (x,y) to the ellipse defined by (x/a)^2 + (y/b)^2 == 1.
*/
/*
double ellipsePerpDist(double x, double y, double a, double b)
{
	double x2 = x*x;
	double y2 = y*y;
	double a4 = a*a*a*a;	
	double a5 = a*a*a*a*a;	
	double a6 = a*a*a*a*a*a;	
	double b4 = b*b*b*b;	
	double b5 = b*b*b*b*b;	
	double b6 = b*b*b*b*b*b;
	
	double AA = sqr(b4*x2 + a4*y2);
	double BB = a*b5*x2 + a5*b*y2;
	double CC = sqrt(b6*x2 + (a6 - sqr(a*a - b*b)*x2)*y2);
	double DD = sqrt(b4*x2 + a4*y2);
	double EE = b6*x2 + a6*y2;
	
	return (AA - BB*CC)/(DD*EE);
}
*/

void hexColour2RGB(const char *val, GLubyte &r, GLubyte &g, GLubyte &b)
{
   if (val[0] != '#')
		ScThrowErr("hex colour must begin with \'#\'");
   if (val[7] != '\0')
		ScThrowErr("hex colour is wrong length");
   unsigned long num;
   sscanf(val+1,"%x",&num);
   //printf("hex num: %x\n", num);
   r = (num >> 16) & 0xFF;
   g = (num >> 8) & 0xFF;
   b = num & 0xFF;
   //printf("r=%x g=%x b=%x\n", r, g, b);
}

Tileset* loadTileset(const char* tilesetFilename, short border, short seperation, bool blend)
{
	char path[256] = "tilesets/";
	char filename[256];
	
	strcat(path, tilesetFilename);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, tilesetFilename);
	strcat(filename, ".xml");
	
	// Load the xml file
	TiXmlDocument *xmlDoc = findXMLFile(filename, gSearchPath);
	if (!xmlDoc)
	{
		ScThrowErr("xml file not found");
	}
	
	// Do the document parsing here
	TiXmlElement *xTileset = xmlDoc->FirstChildElement("tileset");
	if (!xTileset)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileset element not found");
	}
	
	TiXmlHandle tilesetHandle(xTileset);
	
	TiXmlText *xImage = tilesetHandle.FirstChild("image").FirstChild().ToText();
	if (!xImage)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml image element not found");
	}
	
	TiXmlText *xTilewidth = tilesetHandle.FirstChild("tilewidth").FirstChild().ToText();
	if (!xTilewidth)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tilewidth element not found");
	}
	
	TiXmlText *xTileheight = tilesetHandle.FirstChild("tileheight").FirstChild().ToText();
	if (!xTileheight)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileheight element not found");
	}
	
	ScPixelData pixelData;
	SDL_Surface* surface;
	Tileset* tilemap;
	int tileWidth = strtol(xTilewidth->Value(), (char **)NULL, 10);
	int tileHeight = strtol(xTileheight->Value(), (char **)NULL, 10);
	
	if (tileWidth <= 0)
		ScThrowErr("tilewidth is invalid string");
	if (tileHeight <= 0)
		ScThrowErr("tileheight is invalid string");
	
   TileGroup road, offroad, wall;
   
	TiXmlElement *xRoad = tilesetHandle.FirstChildElement("road").Element();
	if (xRoad == NULL)
		ScThrowErr("Could not load road element");
	TiXmlElement *xOffroad = tilesetHandle.FirstChildElement("offroad").Element();
	if (xOffroad == NULL)
		ScThrowErr("Could not load offroad element");
	TiXmlElement *xWall = tilesetHandle.FirstChildElement("wall").Element();
	if (xWall == NULL)
		ScThrowErr("Could not load wall element");
	
   const char *val;
   if (xRoad->QueryIntAttribute("xmin", &road.minx) != TIXML_SUCCESS)
      ScThrowErr("xml road element: xmin attribute not set");    
   if (xRoad->QueryIntAttribute("xmax", &road.maxx) != TIXML_SUCCESS)
      ScThrowErr("xml road element: xmax attribute not set");    
   if (xRoad->QueryIntAttribute("ymin", &road.miny) != TIXML_SUCCESS)
      ScThrowErr("xml road element: ymin attribute not set");    
   if (xRoad->QueryIntAttribute("ymax", &road.maxy) != TIXML_SUCCESS)
      ScThrowErr("xml road element: ymax attribute not set");    
   if (xRoad->QueryDoubleAttribute("grip", &road.grip) != TIXML_SUCCESS)
      ScThrowErr("xml road element: grip attribute not set");    
	if ((val = xRoad->Attribute("dust")) == NULL)
      ScThrowErr("xml road element: dust attribute not set");    
   hexColour2RGB(val, road.r, road.g, road.b);
   
   if (xOffroad->QueryIntAttribute("xmin", &offroad.minx) != TIXML_SUCCESS)
      ScThrowErr("xml offroad element: xmin attribute not set");    
   if (xOffroad->QueryIntAttribute("xmax", &offroad.maxx) != TIXML_SUCCESS)
      ScThrowErr("xml offroad element: xmax attribute not set");    
   if (xOffroad->QueryIntAttribute("ymin", &offroad.miny) != TIXML_SUCCESS)
      ScThrowErr("xml offroad element: ymin attribute not set");    
   if (xOffroad->QueryIntAttribute("ymax", &offroad.maxy) != TIXML_SUCCESS)
      ScThrowErr("xml offroad element: ymax attribute not set");    
   if (xOffroad->QueryDoubleAttribute("grip", &offroad.grip) != TIXML_SUCCESS)
      ScThrowErr("xml offroad element: grip attribute not set");    
	if ((val = xOffroad->Attribute("dust")) == NULL)
      ScThrowErr("xml road element: dust attribute not set");    
   hexColour2RGB(val, offroad.r, offroad.g, offroad.b);
	
   if (xWall->QueryIntAttribute("xmin", &wall.minx) != TIXML_SUCCESS)
      ScThrowErr("xml wall element: xmin attribute not set");    
   if (xWall->QueryIntAttribute("xmax", &wall.maxx) != TIXML_SUCCESS)
      ScThrowErr("xml wall element: xmax attribute not set");    
   if (xWall->QueryIntAttribute("ymin", &wall.miny) != TIXML_SUCCESS)
      ScThrowErr("xml wall element: ymin attribute not set");    
   if (xWall->QueryIntAttribute("ymax", &wall.maxy) != TIXML_SUCCESS)
      ScThrowErr("xml wall element: ymax attribute not set");    
   if (xWall->QueryDoubleAttribute("grip", &wall.grip) != TIXML_SUCCESS)
      ScThrowErr("xml wall element: grip attribute not set");    
	if ((val = xWall->Attribute("dust")) == NULL)
      ScThrowErr("xml road element: dust attribute not set");    
   hexColour2RGB(val, wall.r, wall.g, wall.b);
   
	// create the tile map
	strcpy(filename, path);
	strcat(filename, xImage->Value());
	surface = IMG_Load(filename);
	if (surface == NULL) {
		fprintf(stderr, "SDL_image error : %s\n", IMG_GetError());
		ScThrowErr("Could not load tiles image file");
   }
	PixelDataForSurface(surface, pixelData);
	tilemap = new Tileset(pixelData, tileWidth, tileWidth, border, seperation, blend, road, offroad, wall);
	SDL_FreeSurface(surface);
   
	printf("successfully created tileset from %s\n", path);
	
	return tilemap;
}

void readPointsFromString(const char* str, Tuple* points, short& count)
{
    char *word;//, *last;
    char *temp = strdup(str);
    
    for (   word = strtok_r(temp, ";", &last), count = 0;
            word;
            word = strtok_r(NULL, ";", &last), count++)
    {
        points[count].x = atof(word);
        if ((word = strchr(word, ',')) == NULL)
            ScThrowErr("Badly formatted points list");
        points[count].y = atof(word + 1);
    }
    
    free(temp);
}

void readIntervalsFromString(const char* str, RoadSignInterval* intervals, short& count)
{
   //    char *word, *last;
   char *word;
   char *temp = strdup(str);

   for ( word = strtok_r(temp, ";", &last), count = 0;
         word;
         word = strtok_r(NULL, ";", &last), count++)
   {
      intervals[count].type = (SignType) atoi(word);
      if ((word = strchr(word, ',')) == NULL)
         ScThrowErr("Badly formatted intervals list");
      word++;
      intervals[count].start = atof(word);
      if ((word = strchr(word, ',')) == NULL)
         ScThrowErr("Badly formatted intervals list");
      word++;
      intervals[count].end = atof(word);
   }

   free(temp);
}

char* chooseToyCar(int playerNumber)
{
   const int kMaxCars = 64;
   char str[256];
	// Load the xml file
	TiXmlDocument *xmlDocCarslist = new TiXmlDocument("cars/carslist.xml");
	if (!xmlDocCarslist->LoadFile())
	{
		delete xmlDocCarslist;
		ScThrowErr("xml carslist file not found");
	}
	
	TiXmlHandle docHandleCarlist(xmlDocCarslist);
	TiXmlElement *xCar = docHandleCarlist.FirstChildElement("carslist").FirstChild("car").Element();
	char *carNames[kMaxCars];
    int numCars;
    
    for (numCars = 0; xCar != NULL && numCars < kMaxCars; numCars++) {
        carNames[numCars] = strdup(xCar->FirstChild()->Value());
        xCar = xCar->NextSiblingElement("car");
    }
    
   // load the cars from the user data directory
   if (gUserDataDir[0] != '\0') {
   try {
      strncpy(str, gUserDataDir, 255 - strlen("cars/carslist.xml"));
      strcat(str, "cars/carslist.xml");
      TiXmlDocument *xmlDocUser = new TiXmlDocument(str);
      if (!xmlDocUser->LoadFile())
      {
         delete xmlDocUser;
         ScThrowErr("xml user carslist file not found");
      }
      TiXmlHandle docHandleUser(xmlDocUser);
      xCar = docHandleUser.FirstChildElement("carslist").FirstChild("car").Element();
       for (; xCar != NULL && numCars < kMaxCars; numCars++) {
           carNames[numCars] = strdup(xCar->FirstChild()->Value());
           xCar = xCar->NextSiblingElement("car");
       }
   }
   catch (ScException e)
   {
      e.printMsg();
      printf("Ignoring exception...\n");
   }
   }

    char menuTitle[256];
    sprintf(menuTitle, "Choose Player %d Car:", playerNumber);
    
    Menu menu(carNames, numCars, menuTitle, kLoadingImageFileName);
    int choice = menu.doMenu();
    if (choice == -1)
      return NULL;
    
	char *returnValue = strdup(carNames[choice]);
	
	for (int i = 0; i < numCars; i++)
		free(carNames[i]);
	
	return returnValue;
}

// checks to see if filename starts with $data and resolves this to the default data directory if so
void resolvePath(char *outPath, const char *currentPath, const char *filename)
{
   bool needsResolve = (strncmp(filename, "$data", 5) == 0);
   if (needsResolve) {
      strcpy(outPath, gDataDir);
      strcat(outPath, "/data");
      strcat(outPath, filename + 5);
   }
   else {
      strcpy(outPath, currentPath);
      strcat(outPath, filename);
   }
}

ToyCar* loadNewToyCar(list<char*> &spriteNames, list<ToyCar*> &cars, char *carName, const Tuple& startLocation, double startOrientation)
{
	int i;
	char path[256];
	char filename[256];
   char foundPath[256];
   
	strcpy(path, "cars/");
	strcat(path, carName);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, carName);
	strcat(filename, ".xml");
    
	// Load the xml file
	TiXmlDocument *xmlDocCar = findXMLFile(filename, gSearchPath, foundPath);
	if (!xmlDocCar)
		ScThrowErr("xml file not found");
    
	TiXmlHandle docHandleCar(xmlDocCar);
	TiXmlElement *xSprite = docHandleCar.FirstChildElement("car").FirstChild("sprite").Element();
    int framewidth, frameheight;
    
    if (xSprite == NULL)
        ScThrowErr("xml sprite element not found");
    if (xSprite->QueryIntAttribute("framewidth", &framewidth) != TIXML_SUCCESS)
		ScThrowErr("xml sprite element: frame width attribute not set");    
    if (xSprite->QueryIntAttribute("frameheight", &frameheight) != TIXML_SUCCESS)
		ScThrowErr("xml sprite element: frame height attribute not set");
    
    TiXmlText *xSpriteFilename = docHandleCar.FirstChildElement("car").FirstChild("sprite").FirstChild().ToText();
    if (xSpriteFilename == NULL)
        ScThrowErr("xml sprite filename not found");
    
    const int maxConvexes = 16;
	TiXmlElement *xConvex = docHandleCar.FirstChildElement("car").FirstChild("geometry").FirstChild("convex").Element();
    Convex* convexes[maxConvexes];
    short convexCount = 0;
    Tuple points[256];
    short pointsCount;
    while (xConvex != NULL)
    {
        if (convexCount >= maxConvexes)
            ScThrowErr("xml geometry: too many convex objects inside geometry");
        readPointsFromString(xConvex->FirstChild()->Value(), points, pointsCount);
		for (i = 0; i < pointsCount; i++)
			points[i] = points[i] / kScale;
        convexes[convexCount] = new Convex(points, pointsCount);
        xConvex = xConvex->NextSiblingElement("convex");
        convexCount++;
    }
    Geom *geom = new Geom(convexes, convexCount);
    double axelwidth, frontaxel, rearaxel, frontwheelinset, frontwheelradius, rearwheelradius, mass, inertia, frontwheelinertia, rearwheelinertia, drag, rollingresistance;
    
    TiXmlElement *xProperties = docHandleCar.FirstChildElement("car").FirstChild("properties").Element();
    if (xProperties == NULL)
        ScThrowErr("xml properties element not found");
    if (!(  xProperties->Attribute("axelwidth", &axelwidth) &&
            xProperties->Attribute("frontaxel", &frontaxel) &&
            xProperties->Attribute("rearaxel", &rearaxel) &&
            xProperties->Attribute("frontwheelinset", &frontwheelinset) &&
            xProperties->Attribute("frontwheelradius", &frontwheelradius) &&
            xProperties->Attribute("rearwheelradius", &rearwheelradius) &&
            xProperties->Attribute("mass", &mass) &&
            xProperties->Attribute("inertia", &inertia) &&
            xProperties->Attribute("frontwheelinertia", &frontwheelinertia) &&
            xProperties->Attribute("rearwheelinertia", &rearwheelinertia) &&
            xProperties->Attribute("drag", &drag) &&
            xProperties->Attribute("rollingresistance", &rollingresistance)))
    {
		ScThrowErr("xml car properties attributes missing");
    }
    
    int wheelframewidth, wheelframeheight;
    char wheelLayer;
    TiXmlElement *xWheelsprite = docHandleCar.FirstChildElement("car").FirstChild("wheelsprite").Element();
    TiXmlText *xWheelspriteFilename = docHandleCar.FirstChildElement("car").FirstChild("wheelsprite").FirstChild().ToText();
    
    if (xWheelsprite == NULL)
        ScThrowErr("xml wheelsprite element not found");
    if (xWheelspriteFilename == NULL)
        ScThrowErr("xml wheelsprite filename not found");
    if (xWheelsprite->QueryIntAttribute("framewidth", &wheelframewidth) != TIXML_SUCCESS)
        ScThrowErr("xml car wheelsprite framewidth missing");
    if (xWheelsprite->QueryIntAttribute("frameheight", &wheelframeheight) != TIXML_SUCCESS)
        ScThrowErr("xml car wheelsprite frameheight missing");
    
    if (xWheelsprite->Attribute("layer") != NULL)
    {
        if (strcmp(xWheelsprite->Attribute("layer"), "above") == 0)
            wheelLayer = 1;
        else if (strcmp(xWheelsprite->Attribute("layer"), "below") == 0)
            wheelLayer = 0;
        else
            ScThrowErr("xml car wheelsprite layer attribute has unknown value");
    }
    else {
        wheelLayer = 0;
    }
    
    TiXmlText *front_SR_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("front_SR_Pacejka").FirstChild().ToText();
    TiXmlText *front_SA_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("front_SA_Pacejka").FirstChild().ToText();
    TiXmlText *rear_SR_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("rear_SR_Pacejka").FirstChild().ToText();
    TiXmlText *rear_SA_Pacejka = docHandleCar.FirstChildElement("car").FirstChild("rear_SA_Pacejka").FirstChild().ToText();
    
    if (front_SR_Pacejka == NULL)
        ScThrowErr("xml car front_SR_Pacejka element not found");
    if (front_SA_Pacejka == NULL)
        ScThrowErr("xml car front_SA_Pacejka element not found");
    if (rear_SR_Pacejka == NULL)
        ScThrowErr("xml car rear_SR_Pacejka element not found");
    if (rear_SA_Pacejka == NULL)
        ScThrowErr("xml car rear_SA_Pacejka element not found");
    
    char *word, *temp; //, *brkt;
    double constants[4];
    
    PacejkaCurve *frontSR, *frontSA, *rearSR, *rearSA;
    
    // front slip ratio Pacejka curve
    temp = strdup(front_SR_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml front SR Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
        ScThrowErr("xml front SR Pacejka Curve: too few constants.");
    frontSR = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
    free(temp);
    
    // front slip angle Pacejka curve
    temp = strdup(front_SA_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml front SA Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
        ScThrowErr("xml front SA Pacejka Curve: too few constants.");
    frontSA = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
    free(temp);
    
    // rear slip ratio Pacejka curve
    temp = strdup(rear_SR_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml rear SR Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
        ScThrowErr("xml rear SR Pacejka Curve: too few constants.");
    rearSR = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
    free(temp);
    
    // rear slip angle Pacejka curve
    temp = strdup(rear_SA_Pacejka->Value());
    for (   word = strtok_r(temp, ",", &brkt), i = 0;
            word;
            word = strtok_r(NULL, ",", &brkt), i++)
    {
        if (i >= 4)
            ScThrowErr("xml rear SA Pacejka Curve: too many constants.");
        constants[i] = atof(word);
    }
    if (i < 4)
      ScThrowErr("xml rear SA Pacejka Curve: too few constants.");
   rearSA = new PacejkaCurve(constants[0], constants[1], constants[2], constants[3]);
   free(temp);
   
   char spritefilename[256];
   char wheelspritefilename[256];
   
   strcat(foundPath, "/");
   strcat(foundPath, path);
   resolvePath(spritefilename, foundPath, xSpriteFilename->Value());
   resolvePath(wheelspritefilename, foundPath, xWheelspriteFilename->Value());
   
   // check to see if sprite is already loaded
   list<char*>::const_iterator nameI;
   list<ToyCar*>::const_iterator carI;
   for (nameI = spriteNames.begin(), carI = cars.begin(); nameI != spriteNames.end(); nameI++, carI++) {
      if (strcmp(*nameI, spritefilename) == 0) {
          //printf("Loading duplicate sprite %s == %s\n", spritefilename, *nameI);
          ToyCar* toycar = new ToyCar(   **carI,
                                          mass,
                                          inertia,
                                          startLocation,
                                          startOrientation,
                                          frontaxel / kScale,
                                          rearaxel / kScale,
                                          axelwidth / kScale,
                                          frontwheelinset / kScale,
                                          frontwheelradius / kScale,
                                          rearwheelradius / kScale,
                                          frontwheelinertia,
                                          rearwheelinertia,
                                          *frontSR,
                                          *frontSA,
                                          *rearSR,
                                          *rearSA,
                                          *geom,
                                          drag,
                                          rollingresistance);
          
          return toycar;
      }
   }
   
   // so this is a new sprite
   
   /*
   strcpy(spritefilename, foundPath);
   strcat(spritefilename, "/");
   strcat(spritefilename, path);
   strcat(spritefilename, xSpriteFilename->Value());
   
   strcpy(wheelspritefilename, foundPath);
   strcat(wheelspritefilename, "/");
   strcat(wheelspritefilename, path);
   strcat(wheelspritefilename, xWheelspriteFilename->Value());
   */
    ToyCar* toycar = new ToyCar(    spritefilename,
                                    framewidth,
                                    frameheight,
                                    
                                    wheelspritefilename,
                                    wheelframewidth,
                                    wheelframeheight,
                                    wheelLayer,
                                    
                                    mass,
                                    inertia,
                                    startLocation,
                                    startOrientation,
                                    frontaxel / kScale,
                                    rearaxel / kScale,
                                    axelwidth / kScale,
                                    frontwheelinset / kScale,
                                    frontwheelradius / kScale,
                                    rearwheelradius / kScale,
                                    frontwheelinertia,
                                    rearwheelinertia,
                                    *frontSR,
                                    *frontSA,
                                    *rearSR,
                                    *rearSA,
                                    *geom,
                                    drag,
                                    rollingresistance);
    
    //printf("loaded sprite %s\n", spritefilename);
    spriteNames.push_back(strdup(spritefilename));
    cars.push_back(toycar);
    
    return toycar;
}

Player** players;
/*
void loadPlayers()
{
	// create the players
	players = new Player * [numPlayers];
	ToyCar *toyCar;
	double xpos, ypos, theta;
	int tilewidth = tilemap->getTileWidth();
	int tileheight = tilemap->getTileHeight();
	
	TiXmlElement *xPlayer = docHandle.FirstChildElement("track").FirstChildElement("player").Element();
	for (i = 0; i < numPlayers; i++)
	{
		if (!xPlayer)
		{
			for (j = 0; j < i; j++)
				delete players[j];
			delete [] players;
			xmlDoc->Clear();
			delete xmlDoc;
			ScThrowErr("xml player element not found");
		}
		if (!(xPlayer->Attribute("xpos", &xpos) && xPlayer->Attribute("ypos", &ypos) && xPlayer->Attribute("theta", &theta)))
		{
			for (j = 0; j < i; j++)
				delete players[j];
			delete [] players;
			xmlDoc->Clear();
			delete xmlDoc;
			ScThrowErr("xml player attributes not set");
		}
        
        // check if the sprite is already loaded and load it if not
        toyCar = NULL;
        for (j = 0; j < i; j++) {
            if (strcmp(carNames[j], carNames[i]) == 0) {
                toyCar = new ToyCar(players[j]->car, Tuple(xpos*tilewidth/kScale, ypos*tileheight/kScale), theta);
                break;
            }
        }
        if (toyCar == NULL)
            toyCar = loadNewToyCar(carNames[i], Tuple(xpos*tilewidth/kScale, ypos*tileheight/kScale), theta);
        
		if (i > 0)
			players[i] = new RobotPlayer(*toyCar, pathPolygon, i+1);
		else
			players[i] = new Player(*toyCar, pathPolygon, i+1);
		xPlayer = xPlayer->NextSiblingElement("player");
	}
}
*/

Tuple computePathLocation(const list<Tuple> *pathPolygon, const double loc) {
   if (pathPolygon->size() == 1)
      return pathPolygon->front();
   if (pathPolygon->empty())
      ScThrowErr("Attempt to compute path location but no path exists.");
   /*
   if (loc < 0)
      printf("warning: negative path loc\n");//ScThrowErr("Attempt to compute path location with negative value.");
   */
   double dist = 0;
   list<Tuple>::const_iterator tmp;
   while (1) {
      for (list<Tuple>::const_iterator it = pathPolygon->begin(); it != pathPolygon->end(); it++) {
         tmp = it;
         tmp++;
         if (tmp == pathPolygon->end())
            tmp = pathPolygon->begin();
         double segmentDist = (*tmp - *it).Length();
         if (dist + segmentDist >= loc) {
            double offset = loc - dist;
            Tuple N = (*tmp - *it).Direction() * offset;
            return *it + N;
         }
         dist += segmentDist;
      }
   }
   
   return Tuple(0,0);
}

/*
Loads and sets up a game for split-screen up to n players, given the name
of the track folder containing an xml file with the same name describing the track.
*/
CarGame* loadRacingGame(const char* trackfilename, FMOD::System* FMOD_system)
{
   unsigned int i, j, numPlayers;
   char *items[] = {"New AI Player", "New Human Player", "Done"};
   char menuStr[256];
   Menu menu(items, 3, menuStr, kLoadingImageFileName);
   int choice;
	list<char*> carNames;
	list<int> playerTypes;
	numPlayers = 0;
	gNumHumanPlayers = 0;
	
	// setup player types and car types
	/*
	if (gGameMode == GameMode_HotPotato)
	{
		numPlayers = 2;
		DrawSplash();
		carNames.push_back(chooseToyCar(1));
		DrawSplash();
		carNames.push_back(chooseToyCar(2));
		playerTypes.push_back(1);
		playerTypes.push_back(1);
		gNumHumanPlayers = 2;
	}
	else*/
	{
		// Display player type selection menus
		while (1)
		{
			//DrawSplash();
			sprintf(menuStr, "Add Player %d", numPlayers+1);
			choice = menu.doMenu();
         if (choice == -1)
            return NULL;
			if (choice == 2) {
				if ((gGameMode == GameMode_TimeTrial && numPlayers >= 1) || numPlayers >= 2)
					break;
				else
					continue;
			}
			else if (choice == 1)
				gNumHumanPlayers++;
			numPlayers++;
			//DrawSplash();
         char* toycar = chooseToyCar(numPlayers);
         if (toycar == NULL)
            return NULL;
			carNames.push_back(toycar);
			playerTypes.push_back(choice);
			if (numPlayers >= kMaxPlayersInGame)
				break;
		}
	}
    gNumPlayers = numPlayers;
	
	// draw loading screen
   DrawSplash();
   DrawLoading();
   
	char path[256] = "tracks/";
	char filename[256];
	char foundPath[256];
   
	strcat(path, trackfilename);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, trackfilename);
	strcat(filename, ".xml");
	
	// Load the xml file
	TiXmlDocument *xmlDoc = findXMLFile(filename, gSearchPath, foundPath);
	if (!xmlDoc)
	{
      ScThrowErr("xml track file not found");
   }
	
	TiXmlHandle docHandle(xmlDoc);
	
	TiXmlElement *xTrack = docHandle.FirstChildElement("track").Element();
	TiXmlText *xTileset = docHandle.FirstChildElement("track").FirstChild("tileset").FirstChild().ToText();
	TiXmlText *xTilemap = docHandle.FirstChildElement("track").FirstChild("tilemap").FirstChild().ToText();
	TiXmlElement *xStartline = docHandle.FirstChildElement("track").FirstChildElement("startline").Element();
	TiXmlText *xPath = docHandle.FirstChildElement("track").FirstChild("path").FirstChild().ToText();
	TiXmlText *xRoadSigns = docHandle.FirstChildElement("track").FirstChild("roadsigns").FirstChild().ToText();
	
	// do error checking
	if (!(xTrack && xTileset && xTilemap && xStartline))
	{
		xmlDoc->Clear();
		delete xmlDoc;
		if (!xTrack)
			ScThrowErr("xml track element not found");
		if (!xTileset)
			ScThrowErr("xml tileset element not found");
		if (!xTilemap)
			ScThrowErr("xml tilemap element not found");
		if (!xStartline)
			ScThrowErr("xml startline element not found");
		if (!xPath)
			ScThrowErr("xml path element not found");
	}
	
	// load and create the tileset from an xml file
	Tileset* tilemap = loadTileset(xTileset->Value());
	
	// load the tilemap from the specified tilemap file
   strcpy(filename, foundPath);
   strcat(filename, "/tracks/");
   strcat(filename, trackfilename);
   strcat(filename, "/");
   strcat(filename, xTilemap->Value());
   tilemap->loadMap(filename);
	
	// the path
	Tuple points[512];
	short pointsCount;
	list<Tuple> *pathPolygon = new list<Tuple>;
    readPointsFromString(xPath->Value(), points, pointsCount);
	for (short k = 0; k < pointsCount; k++) {
		pathPolygon->push_back(points[k] * tilemap->getTileWidth() / kScale);
		//printf("%lf, %lf; ", points[k].x, points[k].y);
	}
	//printf("\n");
   
	// the roadsigns
	list<RoadSignInterval> *signIntervals = new list<RoadSignInterval>;
	signIntervals->clear();
	if (xRoadSigns != NULL) {
		RoadSignInterval intervals[256];
		short count;
		readIntervalsFromString(xRoadSigns->Value(), intervals, count);
		for (short k = 0; k < count; k++) {
         intervals[k].startPoint = computePathLocation(pathPolygon, intervals[k].start);
         intervals[k].endPoint = computePathLocation(pathPolygon, intervals[k].end);
			signIntervals->push_back(intervals[k]);
		}
	}
	
	// create the players
	players = new Player * [numPlayers];
	ToyCar *toyCar;
	double xpos, ypos, theta;
	int tilewidth = tilemap->getTileWidth();
	int tileheight = tilemap->getTileHeight();
	list<char*>::const_iterator carNamesIterator = carNames.begin();
	list<char*>::const_iterator otherCarNamesIterator = carNames.begin();
   list<int>::const_iterator playerTypesIterator = playerTypes.begin();
	
   list<char*> spriteNames;
   list<ToyCar*> cars;
   
	xpos = 0;
	ypos = 0;
	theta = 0;
	for (i = 0; i < numPlayers; i++, carNamesIterator++, playerTypesIterator++)
	{
        // check if the sprite is already loaded and load it if not
        toyCar = NULL;
        for (otherCarNamesIterator = carNames.begin(), j = 0; otherCarNamesIterator != carNamesIterator; otherCarNamesIterator++, j++) {
            if (strcmp(*otherCarNamesIterator, *carNamesIterator) == 0) {
                toyCar = new ToyCar(players[j]->car, Tuple(xpos,ypos), theta);
                break;
            }
        }
      if (toyCar == NULL)
         toyCar = loadNewToyCar(spriteNames, cars, *carNamesIterator, Tuple(xpos,ypos), theta);
		if (*playerTypesIterator == 0) {
			if (gGameMode == GameMode_HotPotato)
				players[i] = new RobotHotPotatoPlayer(*toyCar, pathPolygon, i+1, FMOD_system);
			else
				players[i] = new RobotPlayer(*toyCar, pathPolygon, i+1, FMOD_system);
		}
		else {
			players[i] = new Player(*toyCar, pathPolygon, i+1, FMOD_system);
		}
	}
   
   // free the sprite names
   for (list<char*>::const_iterator nameI = spriteNames.begin(); nameI != spriteNames.end(); nameI++) {
      free(*nameI);
   }
	
	// get the start line attributes
	Tuple p1, p2;
   const char *directionStr;
   bool clockwise;
	if (!(xStartline->Attribute("x1", &p1.x) && xStartline->Attribute("y1", &p1.y) &&
			xStartline->Attribute("x2", &p2.x) && xStartline->Attribute("y2", &p2.y) && (directionStr = xStartline->Attribute("direction"))))
	{
		for (j = 0; j < numPlayers; j++)
			delete players[j];
		delete [] players;
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml startline attributes not set");
	}
	p1.x *= tilewidth;
	p2.x *= tilewidth;
	p1.y *= tileheight;
	p2.y *= tileheight;
	if (strcmp(directionStr, "clockwise") == 0) {
      clockwise = true;
   }
   else if (strcmp(directionStr, "anti-clockwise") == 0) {
      clockwise = false;
   }
   else {
      ScThrowErr("xml startline attribute direction is invalid");
   }
	
	// create the game
	CarGame* game = NULL;
	switch (gGameMode) {
		case GameMode_TimeTrial:
			game = new CarGame(tilemap, players, numPlayers, FMOD_system, p1, p2, clockwise, pathPolygon, signIntervals);
			break;
		case GameMode_HotPotato:
			game = new HotPotatoGame(tilemap, players, numPlayers, FMOD_system, p1, p2, clockwise, gPreferences->getFuseLength(), pathPolygon, signIntervals);
			break;
		case GameMode_KnockEmOut:
			game = new KnockOutGame(tilemap, players, numPlayers, FMOD_system, p1, p2, clockwise, pathPolygon, signIntervals);
			break;
		default:
			printf("error: unknown game mode received\n");
			exit(EXIT_FAILURE);
	}
	if (!game)
	{
		for (j = 0; j < i; j++)
			delete players[j];
		delete [] players;
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("failed to create the game object");
	}
	
	// Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
	
	printf("successfully created track from %s\n", path);
	
	// add views to the game
	if ((gGameMode == GameMode_HotPotato || gGameMode == GameMode_TimeTrial) && numPlayers >= 2)
	{
      if (playerTypes.front() == 1 && *(++playerTypes.begin()) == 1) {
         // if first two players are both human controlled use split screen for them
         ScPortRect screenRect = {0, 0, gPreferences->getScreenWidth()/2 - 1, gPreferences->getScreenHeight()};
         game->addView(screenRect, 1);
         screenRect.left = gPreferences->getScreenWidth()/2 + 0;
         game->addView(screenRect, 0);
      }
	}
	if (game->getViews().empty())
	{
		ScPortRect screenRect = {0, 0, gPreferences->getScreenWidth(), gPreferences->getScreenHeight()};
		game->addView(screenRect, 0);
	}
	
	printf("Loading all textures...\n");
	
	// pre-draws all the textures loading them into the vram now
	game->renderEveryTexture();
	
	// clear the screen
	glClear(GL_COLOR_BUFFER_BIT);
	SDL_GL_SwapBuffers();
	glClear(GL_COLOR_BUFFER_BIT);
    
	return game;
}

void InitGL(int width, int height)	        // we call this right after our OpenGL window is created.
{
	// Set up OpenGL for 2D rendering.
	
	glViewport(0, 0, width, height);
   //glScissor(0, 0, width, height);
	glClearColor(0,0,0,1);
   //glClearAccum(0,0,0,1);
   glClear(GL_COLOR_BUFFER_BIT);
   //glClear(GL_ACCUM_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4ub(255, 255, 255, 255);
   glEnable(GL_POLYGON_SMOOTH);
}

void DrawGLScene()
{
	// nb: potential bug here caused by starting a game for the second time without reinitialising static vars
	static double lastTime = GetSeconds();
	static double thisTime;
	
	// the following line of code seems to be not a good idea, so I'll get rid of it by commenting it out
	//while ((thisTime = GetSeconds()) - lastTime < 1e-9) ;	// slow down on very fast machines
	
	thisTime = GetSeconds();
	gGameP->step(std::min(thisTime - lastTime, 0.02));
	lastTime = thisTime;
	/*
   const float q = 0.6;
   glAccum(GL_MULT, q);
   glAccum(GL_ACCUM, 1-q);
   glAccum(GL_RETURN, 1.0);*/
   
	// swap buffers to display, since we're double buffered.
   glFlush();
	SDL_GL_SwapBuffers();
}

void PauseGame()
{
   SDL_Event event;
   int playerNum;
   JoystickButton btn;

   printf("paused... press p/start to unpause\n");
   DrawPaused();
   while (1) {
      if (SDL_WaitEvent(&event) == 0) {
         printf("error waiting for events\n");
         exit(EXIT_FAILURE);
      }
      if (event.key.keysym.sym == SDLK_p && event.key.type == SDL_KEYUP)
         break;
	  if (event.key.type == SDL_JOYBUTTONUP) {
         playerNum = mapJoystickDeviceToPlayer(event.jaxis.which);
         btn = mapJoyButtonInput(event.jbutton.button, gPreferences->getControllerType(playerNum));
         if (btn == kStartButton)
			 break;
	  }
   }
   while (1) {
      SDL_PollEvent(&event);
      if (event.key.keysym.sym == SDLK_p && event.key.type == SDL_KEYDOWN)
         break;
	  if (event.key.type == SDL_JOYBUTTONDOWN) {
         playerNum = mapJoystickDeviceToPlayer(event.jaxis.which);
         btn = mapJoyButtonInput(event.jbutton.button, gPreferences->getControllerType(playerNum));
         if (btn == kStartButton)
			 break;
	  }
   }
}

void MainLoop(bool autoAccelerate)
{
	bool running;
	SDL_Event event;
	unsigned long startTime, frames;
	//#if GAMEPAD_CONTROL_PLAYER_1
   SDL_Joystick *joystick;
	bool joyReverseBtn[4] = {false,false,false,false};
	int numJoysticksRequired = 0;
   int i, playerNum;
   JoystickButton btn;
   JoystickAxis axis;
   short axisValue;
   
   for (i = 0; i < 4; i++)
      numJoysticksRequired += gPreferences->getControllerType(i) >= kXBox360Controller ? 1 : 0;
   
	printf("Number of joysticks available: %d\n", SDL_NumJoysticks());
	SDL_JoystickEventState(SDL_ENABLE);
	joystick = SDL_JoystickOpen(0);
	//#endif
	
	//printf("autoAccelerate is %d\n", autoAccelerate);
	
	if (autoAccelerate)
	{
		// Start with throttle key in down position
		PlayerEventKey playerEvent;
		playerEvent.type = kPlayerKeyDown;
		playerEvent.key = kThrottleMore;
		for (unsigned int i = 0; i < gNumPlayers; i++)
		{
			players[i]->doEvent(&playerEvent);
		}
	}
		
	// Loop, drawing and checking events
	running = true;
	frames = 0;
	startTime = SDL_GetTicks();
	while (running)
	{
		// get pending events
		while (SDL_PollEvent(&event))
		{
			if (event.type == SDL_QUIT)
				running = false;
			
			// checks to see if user clicked on a player and print out debugging info on that player if so
			if (event.type == SDL_MOUSEBUTTONDOWN)
			{
				const unsigned short xloc = event.button.x;
				const unsigned short yloc = gPreferences->getScreenHeight() - event.button.y;
				for (list<ViewFollow>::iterator i = gGameP->getViews().begin(); i != gGameP->getViews().end(); i++) {
					for (unsigned int n = 0; n < gNumPlayers; n++) {
						ScPortRect r = i->view->getPortRect();
						if (r.left < xloc && xloc < r.left + r.width && r.bottom < yloc && yloc < r.bottom + r.height) {
							Tuple pos = players[n]->car.getLocation_wc()*kScale - Tuple(i->view->getViewRect().left,i->view->getViewRect().bottom);
							if ((pos - Tuple(xloc - r.left,yloc - (gPreferences->getScreenHeight() - (r.bottom + r.height)))).Norm() < 400) {
								players[n]->printDebugInfo();
							}
						}
					}
				}
			}
			
			//#if GAMEPAD_CONTROL_PLAYER_1
			if (event.type == SDL_JOYAXISMOTION) {
				playerNum = mapJoystickDeviceToPlayer(event.jaxis.which);
				axis = mapJoyAxisInput(event.jaxis.axis, gPreferences->getControllerType(playerNum), event.jaxis.value, &axisValue);
				PlayerEventPaddle playerEvent;
				playerEvent.type = kPlayerPaddleMove;
				//printf("axis %d: %d\n", event.jaxis.axis, event.jaxis.value);
				if (axis == kLeftHorizontalAxis)
				{
					const double kThresh = 5000;
					playerEvent.control = kSteerWheel;
					if (axisValue > -kThresh && axisValue < kThresh)
						playerEvent.value = 0;
					else
						playerEvent.value = -short((32768.0/(32768.0 - kThresh))*(axisValue > 0 ? axisValue - kThresh : axisValue + kThresh)) - 1;
					gGameP->doEvent(&playerEvent, playerNum-1);
				}
				
				if (axis == kRightTrigger)	// throttle
				{
					playerEvent.control = kThrottle;
					playerEvent.value = joyReverseBtn[playerNum-1] ? -axisValue/2 - 16385 : axisValue/2 + 16384;
					gGameP->doEvent(&playerEvent, playerNum-1);
				}
				
            if (axis == kLeftTrigger)	// brake
				{
					playerEvent.control = kBrake;
					playerEvent.value = axisValue;
					gGameP->doEvent(&playerEvent, playerNum-1);
				}
				/*
				if (axis == kLeftVerticalAxis && axisValue >= 0)	// brake
				{
					axisValue = 2*axisValue - 32767;
					playerEvent.control = kBrake;
					playerEvent.value = axisValue;
					gGameP->doEvent(&playerEvent, playerNum-1);
				}*/
			}
			if (event.type == SDL_JOYHATMOTION) {
				//printf("hat %d: %d\n", event.jhat.hat, event.jhat.value);
         }
			if (event.type == SDL_JOYBUTTONDOWN) {
				playerNum = mapJoystickDeviceToPlayer(event.jbutton.which);
				btn = mapJoyButtonInput(event.jbutton.button, gPreferences->getControllerType(playerNum));
				//printf("joy which: %d button: %d\n", event.jaxis.which, event.jbutton.button);
				if (btn == kAButton) {
					joyReverseBtn[playerNum-1] = true;
				}
				if (btn == kBackButton) {
					running = false;
				}
				if (btn == kBButton) {
					gGameP->changeCameraMode();
				}
				// pause the game
				if (btn == kStartButton) {
               PauseGame();
               continue;
            }
			}
			if (event.type == SDL_JOYBUTTONUP) {
				playerNum = mapJoystickDeviceToPlayer(event.jbutton.which);
				btn = mapJoyButtonInput(event.jbutton.button, gPreferences->getControllerType(playerNum));
				if (btn == kAButton) {
					joyReverseBtn[playerNum-1] = false;
				}
				/*
				if (btn == kLeftButton) {
					PlayerEventKey playerEvent;
					playerEvent.type = kPlayerKeyUp;
					playerEvent.key = kSteerWheelLeft;
					gGameP->doEvent(&playerEvent, playerNum-1);
				}
				if (btn == kRightButton) {
					PlayerEventKey playerEvent;
					playerEvent.type = kPlayerKeyUp;
					playerEvent.key = kSteerWheelRight;
					gGameP->doEvent(&playerEvent, playerNum-1);
				}*/
			}
			//#endif
			
			//#if MOUSE_CONTROL_PLAYER_1
         /*
			if (event.type == SDL_MOUSEMOTION)
			{
            // nb: mouse controlling is buggy
            
				PlayerEventPaddle playerEvent;
				playerEvent.type = kPlayerPaddleMove;
				
				playerEvent.control = kSteerWheel;
				playerEvent.value -= event.motion.xrel * 50;
				gGameP->doEvent(&playerEvent, 0);
				
				playerEvent.control = kThrottle;
				playerEvent.value -= event.motion.yrel * 50;
				gGameP->doEvent(&playerEvent, 0);
			}
         */
			//#endif
			
			if (event.type == SDL_KEYDOWN)
			{
				if (event.key.keysym.sym == SDLK_ESCAPE)
					running = false;
				
				if (event.key.keysym.sym == SDLK_TAB) {
					if (gGameMode == GameMode_HotPotato)
						dynamic_cast<HotPotatoGame*>(gGameP)->getPlayerWithBomb()->printDebugInfo();
				}
				
				if (event.key.keysym.sym == SDLK_f)
				{
					printf("secs/frame = %f\n", (SDL_GetTicks() - startTime) * 1e-3 / frames);
					startTime = SDL_GetTicks();
					frames = 0;
               //printf("slip player 1: %f\n", players[0]->car.getWheel_BL().getLongtitudinalSlipVelocity() + players[0]->car.getWheel_BR().getLongtitudinalSlipVelocity());
				}
				
				// pause the game
				if (event.key.keysym.sym == SDLK_p) {
               PauseGame();
               continue;
            }
            
				if (event.key.keysym.sym == SDLK_c) {
               gGameP->changeCameraMode();
            }
			}
			if (event.type == SDL_KEYDOWN || event.type == SDL_KEYUP)
			{
				PlayerEventKey playerEvent;
				
				switch (event.type) {
					case SDL_KEYDOWN:
						playerEvent.type = kPlayerKeyDown;
						break;
					case SDL_KEYUP:
						playerEvent.type = kPlayerKeyUp;
						switch (event.key.keysym.sym) {
							case SDLK_1:
								if (gNumPlayers < 1)	break;
								printf("Player 1 path distance: %lf.\n", players[0]->getPathDistance());
								printf("Player 1 path location: %lf %lf.\n", players[0]->getPathLocation().x, players[0]->getPathLocation().y);
								break;
							case SDLK_2:
								if (gNumPlayers < 2)	break;
								printf("Player 2 path distance: %lf.\n", players[1]->getPathDistance());
								printf("Player 2 path location: %lf %lf.\n", players[1]->getPathLocation().x, players[1]->getPathLocation().y);
								break;
							case SDLK_3:
								if (gNumPlayers < 3)	break;
								printf("Player 3 path distance: %lf.\n", players[2]->getPathDistance());
								printf("Player 3 path location: %lf %lf.\n", players[2]->getPathLocation().x, players[2]->getPathLocation().y);
								break;
							case SDLK_4:
								if (gNumPlayers < 4)	break;
								printf("Player 4 path distance: %lf.\n", players[3]->getPathDistance());
								printf("Player 4 path location: %lf %lf.\n", players[3]->getPathLocation().x, players[3]->getPathLocation().y);
								break;
							default:
								break;
						}
						break;
				}
				
				switch (event.key.keysym.sym)
				{
					
				// player 0 keys
				//#if !(MOUSE_CONTROL_PLAYER_1 || GAMEPAD_CONTROL_PLAYER_1)
					case SDLK_LEFT:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 0);
						break;
					case SDLK_RIGHT:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 0);
						break;
					case SDLK_DOWN:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 0);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 0);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 0);
							}
						}
						break;
					case SDLK_UP:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 0);
						break;
				//#endif
				
				// player 1 keys
				
					case SDLK_a:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 1);
						break;
					case SDLK_d:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 1);
						break;
					case SDLK_s:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 1);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 1);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 1);
							}
						}
						break;
					case SDLK_w:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 1);
						break;
				
				// player 2 keys
				
					case SDLK_j:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 2);
						break;
					case SDLK_l:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 2);
						break;
					case SDLK_k:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 2);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 2);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 2);
							}
						}
						break;
					case SDLK_i:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 2);
						break;
				
				// player 3 keys
				
					case SDLK_KP4:
						playerEvent.key = kSteerWheelLeft;
						gGameP->doEvent(&playerEvent, 3);
						break;
					case SDLK_KP6:
						playerEvent.key = kSteerWheelRight;
						gGameP->doEvent(&playerEvent, 3);
						break;
					case SDLK_KP5:
						playerEvent.key = kThrottleLess;
						gGameP->doEvent(&playerEvent, 3);
						if (autoAccelerate) {
							if (event.type == SDL_KEYUP) {
								playerEvent.type = kPlayerKeyDown;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 3);
							}
							else if (event.type == SDL_KEYDOWN) {
								playerEvent.type = kPlayerKeyUp;
								playerEvent.key = kThrottleMore;
								gGameP->doEvent(&playerEvent, 3);
							}
						}
						break;
					case SDLK_KP8:
						if (autoAccelerate)	break;
						playerEvent.key = kThrottleMore;
						gGameP->doEvent(&playerEvent, 3);
						break;
					default:
						break;
				}
			}
		}
		
		// Draw
		DrawGLScene();
		frames++;
	}
	
	//#if GAMEPAD_CONTROL_PLAYER_1
   SDL_JoystickClose(joystick);
	//#endif
}

char chooseTrack(char result[])
{
   const int kMaxTracks = 64;
   char *trackNames[kMaxTracks];
   char str[256];
   
   // load the tracks from the default data directory
	TiXmlDocument *xmlDoc = new TiXmlDocument("tracks/tracklist.xml");
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml tracklist file not found");
	}
	TiXmlHandle docHandle(xmlDoc);
	TiXmlElement *xTrack = docHandle.FirstChildElement("tracklist").FirstChild("track").Element();
   int numTracks;
   for (numTracks = 0; xTrack != NULL && numTracks < kMaxTracks; numTracks++) {
      trackNames[numTracks] = strdup(xTrack->Attribute("name"));
      xTrack = xTrack->NextSiblingElement("track");
   }
   
   // load the tracks from the user data directory
   if (gUserDataDir[0] != '\0') {
   try {
      strncpy(str, gUserDataDir, 255 - strlen("/tracks/tracklist.xml"));
      strcat(str, "/tracks/tracklist.xml");
      TiXmlDocument *xmlDocUser = new TiXmlDocument(str);
      if (!xmlDocUser->LoadFile())
      {
         delete xmlDocUser;
         ScThrowErr("xml user tracklist file not found");
      }
      TiXmlHandle docHandleUser(xmlDocUser);
      xTrack = docHandleUser.FirstChildElement("tracklist").FirstChild("track").Element();
      for (; xTrack != NULL && numTracks < kMaxTracks; numTracks++) {
         trackNames[numTracks] = strdup(xTrack->Attribute("name"));
         xTrack = xTrack->NextSiblingElement("track");
      }
   }
   catch (ScException e)
   {
      e.printMsg();
      printf("Ignoring exception...\n");
   }
   }
   
   // show menu
   Menu menu(trackNames, numTracks, "Choose track:", kLoadingImageFileName);
   int choice = menu.doMenu();
   if (choice == -1)
      return 0;
   strcpy(result, trackNames[choice]);
   for (int i = 0; i < numTracks; i++)
      free(trackNames[i]);
   return 1;
}

void testRand();

int main(int argc, char **argv) 
{
	unsigned long flags;
   
	// Set flags
	flags = SDL_OPENGL;
	
	// Handle command line options
	for (int i = 1; argc > 1; i++, argc--)
	{
		if(strncmp(argv[i], "-datadir", 8) == 0)
		{
			if (argv[i][8] != '=')
			{
				fprintf(stderr, "error: missing argument to -datadir\n");
				exit(EXIT_FAILURE);
			}
			if (strlen(&argv[i][9]) > 255)
			{
				fprintf(stderr, "error: directory path cannot be more than 255 characters long for -datadir\n");
				exit(EXIT_FAILURE);
			}
			strncpy(gDataDir, &argv[i][9], 255);
		}
		else if(strcmp(argv[i], "-h") == 0)
		{
			printf("Usage: %s [OPTION]\n\n", argv[0]);
			printf("Options:\n");
			printf("   -datadir=DIR   Specify the location of the game data directory\n");
			printf("   -h             Display this help text\n");
			exit(0);
		}
		else
		{
			fprintf(stderr, "warning: unrecognised argument \"%s\"\n", argv[i]);
		}
	}
   
#ifdef XCODE
   // in the Xcode bundle DATA_DIR is only a relative directory so we need to make it absolute here.
   getcwd(gDataDir, 255 - strlen(DATA_DIR) - 1);
   strcat(gDataDir, "/");
   strcat(gDataDir, DATA_DIR);
#endif
   
   printf("Default Data Dir: %s\n", gDataDir);

#ifndef WIN32
   if (gDataDir[0] != '/') {
	   fprintf(stderr, "Error: The Default data directory \"%s\" is not an absolute path. It must begin with the '/' character.\n", gDataDir);
      exit(1);
   }
#endif

#ifdef WIN32
   gUserDataDir[0] = '\0';
#endif
    // Load preferences
    try {
    char str[256];
    strcpy(str, gDataDir);
    strcat(str, "/");
    strcat(str, "data");
	
   // change to user data directory
	printf("Changing to directory: %s\n", gDataDir);
	if (chdir(gDataDir) == -1) {
		printf("chdir to directory \"%s\" failed.\n", gDataDir);
		exit(EXIT_FAILURE);
	}
	if (chdir("data") == -1) {
		printf("chdir to directory \"data\" failed.\n");
		exit(EXIT_FAILURE);
	}
   
#ifdef WIN32
   gSearchPath = makeSearchDirectories(gDataDir, NULL);
	gPreferences = new TcPreferences("toycars_prefs.xml");
#else
   // setup the user data dir
    sprintf(gUserDataDir, "%s/.toycars/", getenv("HOME"));
    printf("User data dir: %s\n", gUserDataDir);
    gSearchPath = makeSearchDirectories(str, makeSearchDirectories(gUserDataDir, NULL));
    
    // Check for user data directory. Create if it does not exist.
    int cwd = open(".", O_RDONLY);
    if (chdir(gUserDataDir) == -1) {
        if (errno == ENOENT) {
            // directory does not exist so create it
            sprintf(str, "mkdir %s", gUserDataDir);
            system(str);
        }
        else {
            // an error of some kind occured
            ScThrowErr("chdir command encountered an error");
        }
    }
    fchdir(cwd);
    close(cwd);
    
    sprintf(str, "%s/preferences.xml", gUserDataDir);
    gPreferences = new TcPreferences(str);
#endif
   
	if (gPreferences->getFullscreen())
		flags |= SDL_FULLSCREEN;
    }
    catch (ScException e)
    {
        e.printMsg();
        exit(EXIT_FAILURE);
    }
    
	// Initialize SDL for video output
	if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK) < 0)
	{
		fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}
	
	// Create an OpenGL screen
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE,8);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE,8);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE,8);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE,8);
	SDL_GL_SetAttribute(SDL_GL_BUFFER_SIZE,32);/*
	SDL_GL_SetAttribute(SDL_GL_ACCUM_RED_SIZE,5);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_BLUE_SIZE,5);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_GREEN_SIZE,5);
	SDL_GL_SetAttribute(SDL_GL_ACCUM_ALPHA_SIZE,5);*/
   SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
   if (SDL_SetVideoMode(gPreferences->getScreenWidth(), gPreferences->getScreenHeight(), 0, flags) == NULL)
	{
		fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
		SDL_Quit();
		exit(EXIT_FAILURE);
	}
	
	// Set the title bar in environments that support it
	SDL_WM_SetCaption("toycars", NULL);
    
	{
		printf("%i joysticks were found.\n\n", SDL_NumJoysticks() );
		printf("The names of the joysticks are:\n");
			
		for(int i=0; i < SDL_NumJoysticks(); i++ ) 
		{
			SDL_Joystick *joystick;
			printf("    %s\n", SDL_JoystickName(i));
			joystick = SDL_JoystickOpen(i);
			printf("Num axes: %d\n", SDL_JoystickNumAxes(joystick));
			printf("Num buttons: %d\n", SDL_JoystickNumButtons(joystick));
			printf("Num balls: %d\n", SDL_JoystickNumBalls(joystick));
			printf("Num hats: %d\n", SDL_JoystickNumHats(joystick));
		}
	}
	
	// Output instructions
	printf("toycars\n");
	printf("Arrows: steering and throttle for player 1\n");
	printf("adws: steering and throttle for player 2\n");
	//printf("Page up and down: zoom view\n");
	//printf("Spacebar: shoot\n");
	printf("f: aproximate seconds per frame\n");
	printf("c: change camera mode\n");
	printf("tab: in Hot Potato Mode - print debug info for player holding the bomb\n");
	printf("p: pause\n");
	printf("Esc: quit\n");
	
#ifdef WIN32
	// startup windows timing services
	TIMECAPS resolution;
	
	if (timeGetDevCaps (&resolution, sizeof (TIMECAPS)) == TIMERR_NOERROR)
	{
		printf ("Minimum supported timing resolution = %d\n", resolution.wPeriodMin);
		printf ("Maximum supported timing resolution = %d\n", resolution.wPeriodMax);
	}
	
	if (timeBeginPeriod (resolution.wPeriodMin) != TIMERR_NOERROR)
	{
		printf("timeBeginPeriod failed\n");
		exit(EXIT_FAILURE);
	}
#endif

	// Seed the random number generator
	srand(time(NULL));

	// Initialise OpenGL
	
	// initialise the glf library
	// must be done before calling DrawSplash
	glfInit();
	int fontID = glfLoadFont("fonts/cricket1.glf");
	if (fontID == GLF_ERROR) {
		printf("glfLoadFont returned GLF_ERROR\n");
        exit(EXIT_FAILURE);
    }
	glfSetCurrentFont(fontID);
    
    int menuResult;
    char trackName[256];
    char *items[] = {"New Game", "Preferences", "Credits", "Quit"};
    Menu menu(items, 4, 0, kLoadingImageFileName);
	bool quitting = false;
	
	// Initialize the FMOD library
    FMOD::System     *FMOD_system = 0;
#if HAVE_FMOD_LIB
    FMOD::Sound      *sound;
    FMOD_RESULT       result;
    unsigned int      version;
    FMOD::Channel    *musicChannel = 0;
	bool	playingMusic;

    result = FMOD::System_Create(&FMOD_system);
    ERRCHECK(result);
	
    result = FMOD_system->getVersion(&version);
    ERRCHECK(result);
	
    if (version < FMOD_VERSION)
    {
        printf("Error!  You are using an old version of FMOD %08x.  This program requires %08x\n", version, FMOD_VERSION);
        return 0;
    }
	
    result = FMOD_system->init(kFMODMaxChannels, FMOD_INIT_NORMAL, 0);
    ERRCHECK(result);
	
    result = FMOD_system->createSound(kTitleSongFileName, (FMOD_MODE)(FMOD_SOFTWARE | FMOD_2D | FMOD_CREATESTREAM), 0, &sound);
    ERRCHECK(result);
#endif
	
    do {
        InitGL(gPreferences->getScreenWidth(), gPreferences->getScreenHeight());
        //DrawSplash();
		
#if HAVE_FMOD_LIB
		// (re)start music
		if (gPreferences->getMusic()) {
			result = musicChannel->isPlaying(&playingMusic);
			if (!playingMusic || result == FMOD_ERR_INVALID_HANDLE) {
				result = FMOD_system->playSound(FMOD_CHANNEL_FREE, sound, false, &musicChannel);
				ERRCHECK(result);
			}
		}
#endif
		
		menuResult = menu.doMenu();
        switch (menuResult)
        {
         case -1:
            quitting = true;
            continue;
			case 0:		// New Game
				{
					char *items[] = {"Timed Race", "Hot Potato (2P)", "Knock 'em Out"};
					Menu menu(items, 3, "Choose Game Type", kLoadingImageFileName);
					//DrawSplash();
					switch(menu.doMenu())
					{
                  case -1:
                     continue;
						case 0:     // Time Trial
							gGameMode = GameMode_TimeTrial;
							break;
						case 1:		// Hot Potato
							gGameMode = GameMode_HotPotato;
							break;
						case 2:     // Knock Em Out
							gGameMode = GameMode_KnockEmOut;
							break;
						default:
							printf("error: unknown menu item in choose game type menu");
							exit(EXIT_FAILURE);
					}
				}
				break;
            case 1:     // Preferences
				gPreferences->drawPrefs();
				if (gPreferences->getModifiedScreenPrefs()) {
					if (SDL_SetVideoMode(gPreferences->getScreenWidth(), gPreferences->getScreenHeight(), 0, SDL_OPENGL | (gPreferences->getFullscreen() ? SDL_FULLSCREEN : 0)) == NULL)
					{
						fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
						SDL_Quit();
						exit(EXIT_FAILURE);
					}
				}
#if HAVE_FMOD_LIB
				if (gPreferences->getMusic() == false) {
					result = musicChannel->isPlaying(&playingMusic);
					if (result != FMOD_ERR_INVALID_HANDLE) {
						ERRCHECK(result);
						if (playingMusic)
							musicChannel->stop();
					}
				}
#endif
				continue;
			
         case 2:     // Credits
				DrawSplash();
				ShowCredits();
				continue;
         case 3:     // Quit
				quitting = true;
            continue;
			default:
				printf("error: unknown menu item in main menu");
				exit(EXIT_FAILURE);
        }
        
        //DrawSplash();
        if (chooseTrack(trackName) == 0)
            continue;
        
        printf("Setting up animation...\n");
        
        // load tiles tilemap and car sprites from track xml file
       	try {
            gGameP = loadRacingGame(trackName, FMOD_system);
            if (gGameP == NULL) {
               continue;
            }
         }
        catch (ScException e)
        {
            e.printMsg();
            SDL_Quit();
            exit(EXIT_FAILURE);
        }

         // Enter the main loop
         SDL_ShowCursor(SDL_DISABLE);
         MainLoop(gNumHumanPlayers > 2);
         SDL_ShowCursor(SDL_ENABLE);

         delete gGameP;
         gGameP = NULL;
    } while (quitting == false);
    
	printf("Quitting...\n");
	
   // Clean up
   glfUnloadFont();
   delete gPreferences;
	
   freeSearchDirectories(gSearchPath);
   
   /*
      Shut down FMOD
   */
#if HAVE_FMOD_LIB
    result = sound->release();
    ERRCHECK(result);
    result = FMOD_system->close();
    ERRCHECK(result);
    result = FMOD_system->release();
    ERRCHECK(result);
#endif
	
	// Quit SDL
	SDL_Quit();

#ifdef WIN32
	if (timeEndPeriod (resolution.wPeriodMin) != TIMERR_NOERROR)
	{
		printf("timeEndPeriod failed\n");
		exit(EXIT_FAILURE);
	}
#endif
	
	return 0;
}
