
#include "TrackView.h"
#include <math.h>
#include <cstring>

#include "ScAnimation.h"
#include "tinyxml.h"
#include "ScException.h"
#include <Fl/Fl_JPEG_Image.H>
#include <Fl/Fl_PNG_Image.H>
#include <Fl/Fl_BMP_Image.H>
#include <Fl/Fl_PNM_Image.H>
#include <Fl/Fl_GIF_Image.H>
#include <Fl/Fl_XPM_Image.H>
#include <FL/Fl_File_Chooser.H>   // fl_alert
#include "GameConstants.h"
#include "TrackEditorUI.h"

#ifdef XCODE
#include <Carbon/Carbon.h>
#endif

#ifndef strtok_r
   #define strtok_r(s,d,p) strtok(s,d)
#endif

const unsigned int kMinimapUpdateLimit = 10;
const int kMainWindowWidth = 1024;
const int kMainWindowHeight = 768;
const int kScrollbarMaxValue = 10000;
const double kVertexSelectMinDist = 10 / kScale;
const double kPathSelectMinDist = 10 / kScale;
const double kIntervalHandleRadius = 8 / kScale;

void drawCircleAt(double x, double y, double r, bool fill = true, int n = 10) {
   const double da = 1.0 / n;
   if (fill)
      glBegin(GL_POLYGON);
   else
      glBegin(GL_LINE_LOOP);
   for (double a = 0; a < 1; a += da) {
      glVertex2d((x + r*cos(2*M_PI*a)) * kScale,
                 (y + r*sin(2*M_PI*a)) * kScale);
   }
   glEnd();
}

void freeSearchDirectories(SearchDirectories *dirs) {
   SearchDirectories *temp;
   while (dirs != NULL) {
      temp = dirs->next;
      free(dirs);
      dirs = temp;
   }
}

SearchDirectories *makeSearchDirectories(const char *path, SearchDirectories *next) {
   SearchDirectories *result = new SearchDirectories;
   strncpy(result->path, path, 255);
   result->next = next;
   return result;
}

TiXmlDocument *findXMLFile(const char *filename, SearchDirectories *dirs, char *outPath = NULL)
{
   const int filenameLen = strlen(filename);
   char str[256];
   TiXmlDocument *xmlDoc;
   while (dirs != NULL) {
      if (strlen(dirs->path) + filenameLen + 1 < 255) {
         strcpy(str, dirs->path);
         strcat(str, "/");
         strcat(str, filename);
         xmlDoc = new TiXmlDocument(str);
         if (xmlDoc->LoadFile()) {
            if (outPath)
               strcpy(outPath, dirs->path);
            return xmlDoc;
         }
         else {
            delete xmlDoc;
         }
         dirs = dirs->next;
      }
   }
   return NULL;
}

//{jpg,png,bmp,pnm,pbm,pgm,ppm}
Fl_Image* loadImage(const char *filename)
{
   char *suffix = strrchr(filename, '.');
   if (suffix == NULL)
      return NULL;
   if (strcmp(suffix,".jpg") == 0)
      return new Fl_JPEG_Image(filename);
   else if (strcmp(suffix,".png") == 0)
      return new Fl_PNG_Image(filename);
   else if (strcmp(suffix,".bmp") == 0)
      return new Fl_BMP_Image(filename);
   else if (strcmp(suffix,".pnm") == 0)
      return new Fl_PNM_Image(filename);
   else if (strcmp(suffix,".pbm") == 0)
      return new Fl_PNM_Image(filename);
   else if (strcmp(suffix,".pgm") == 0)
      return new Fl_PNM_Image(filename);
   else if (strcmp(suffix,".ppm") == 0)
      return new Fl_PNM_Image(filename);/*
   else if (strcmp(suffix,".gif") == 0)
      return new Fl_GIF_Image(filename);
   else if (strcmp(suffix,".xpm") == 0)
      return new Fl_XPM_Image(filename);*/
   else
      return NULL;
}

void PixelDataForImage(const Fl_Image* src, ScPixelData& dst)
{
	dst.bytes_pp = src->d();
	dst.width = src->w();
	dst.height = src->h();
	dst.pitch = dst.width * dst.bytes_pp;
	
	switch (dst.bytes_pp)
	{
		case 4:
			dst.format = GL_RGBA;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGBA;
			break;
		
		case 3:
			dst.format = GL_RGB;
			dst.type = GL_UNSIGNED_BYTE;
			dst.internalformat = GL_RGB;
			break;
		
		default:
			ScThrowErr("Only 32 bit or 24 bit is supported at present");
			break;
	}
	
   if (src->count() == 1)
      dst.pixels = src->data()[0];
   else
      ScThrowErr("Unsupported image type");
}

TrackView::TrackView(int x,int y,int w,int h,const char *l)
   : Fl_Gl_Window(x,y,w,h,l), startLine(0), signs(0), view(0), searchPath(0), tiles(0), clockwise(false),
     hoverPathVertexExists(false), selectedPathVertexExists(false), hoverPathIntermediateVertexExists(false), drawingSignInterval(false),
     signIntervalStartMouse(NULL), signIntervalEndMouse(NULL)
{
#ifdef WIN32
   strncpy(dataDir, DATA_DIR, 1023);
   strncpy(userDataDir, DATA_DIR, 1023);
   searchPath = makeSearchDirectories(dataDir, NULL);
#else
 #ifdef XCODE
   // in the Xcode bundle DATA_DIR is only a relative directory so we need to make it absolute here.
   Boolean succ = CFURLGetFileSystemRepresentation(CFBundleCopyBundleURL(CFBundleGetMainBundle()), TRUE, (UInt8*)dataDir, 1023);
   //getcwd(dataDir, 1023 - strlen(DATA_DIR) - 1);
   if (!succ) {
      fprintf(stderr, "Error: Failed to get application bundle path\n");
      exit(1);
   }
   char *s = strrchr(dataDir, '/');
   s[1] = '\0';
   strcat(dataDir, DATA_DIR);
 #else
   strncpy(dataDir, DATA_DIR, 1023);
 #endif // XCODE
   strcat(dataDir,"/data");
   sprintf(userDataDir, "%s/.toycars", getenv("HOME"));
   searchPath = makeSearchDirectories(dataDir, makeSearchDirectories(userDataDir, NULL));
#endif // WIN32
   printf("dataDir = %s\nuserDataDir = %s\n", dataDir, userDataDir);
   
	TiXmlDocument *xmlDoc = findXMLFile("tilesets/default/default.xml", searchPath);
   if (!xmlDoc) {
      fl_alert("Could not find default tileset. Make sure we are in the same directory as the toycars app bundle.\n");
      exit(1);
   }
   delete xmlDoc;
}

TrackView::~TrackView()
{
   if (view)
      delete view;
   if (searchPath)
      freeSearchDirectories(searchPath);
   if (signs)
      delete signs;
}

void TrackView::setTool(ToolType t)
{
   tool = t;
   if (tool != kPathTool) {
      hoverPathVertexExists = false;
      selectedPathVertexExists = false;
      hoverPathIntermediateVertexExists = false;
   }
   redraw();
}

ScTilemap* TrackView::loadTileset(const char* tilesetFilename, short border, short seperation, bool blend)
{
	char path[1024] = "tilesets/";
	char filename[1024];
   char foundPath[1024];
	
	strcat(path, tilesetFilename);
	strcat(path, "/");
	strcpy(filename, path);
	strcat(filename, tilesetFilename);
	strcat(filename, ".xml");
	
	// Load the xml file
	TiXmlDocument *xmlDoc = findXMLFile(filename, searchPath, foundPath);
	if (!xmlDoc)
	{
		ScThrowErr("xml file not found");
	}
   strcat(foundPath, "/");
   
	// Do the document parsing here
	TiXmlElement *xTileset = xmlDoc->FirstChildElement("tileset");
	if (!xTileset)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileset element not found");
	}
	
	TiXmlHandle tilesetHandle(xTileset);
	
	TiXmlText *xImage = tilesetHandle.FirstChild("image").FirstChild().ToText();
	if (!xImage)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml image element not found");
	}
	
	TiXmlText *xTilewidth = tilesetHandle.FirstChild("tilewidth").FirstChild().ToText();
	if (!xTilewidth)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tilewidth element not found");
	}
	
	TiXmlText *xTileheight = tilesetHandle.FirstChild("tileheight").FirstChild().ToText();
	if (!xTileheight)
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml tileheight element not found");
	}
	
	ScPixelData pixelData;
	Fl_Image* img;
	ScTilemap* tilemap;
	int tileWidth = strtol(xTilewidth->Value(), (char **)NULL, 10);
	int tileHeight = strtol(xTileheight->Value(), (char **)NULL, 10);
	
	if (tileWidth <= 0)
		ScThrowErr("tilewidth is invalid string");
	if (tileHeight <= 0)
		ScThrowErr("tileheight is invalid string");
	
	// create the tile map
	strcpy(filename, foundPath);
	strcat(filename, path);
	strcat(filename, xImage->Value());
	img = loadImage(filename);
	if (img == NULL)
		ScThrowErr("Could not load tiles image file");
	PixelDataForImage(img, pixelData);
	tilemap = new ScTilemap(pixelData, tileWidth, tileWidth, border, seperation, blend);
	delete img;
	
	printf("successfully created tileset from %s\n", filename);
	
	xmlDoc->Clear();
   delete xmlDoc;
	return tilemap;
}

void TrackView::resetScroll()
{
   view->scrollTo(0,0);
   boundScroll();
   updateScrollBars();
   redraw();
}

void TrackView::newMap(int w, int h)
{
	tiles->newMap(w, h, kWallTile);
   pathPolygon.clear();
   redraw();
   ui->setChanged(false);
   resetScroll();
}

void readPointsFromString(const char* str, Tuple* points, short& count)
{
//    char *word, *last;
    char *word;
    char *temp = strdup(str);
    
    for (   word = strtok_r(temp, ";", &last), count = 0;
            word;
            word = strtok_r(NULL, ";", &last), count++)
    {
        points[count].x = atof(word);
        if ((word = strchr(word, ',')) == NULL)
            ScThrowErr("Badly formatted points list");
        points[count].y = atof(word + 1);
    }
    
    free(temp);
}

void readIntervalsFromString(const char* str, RoadSignInterval* intervals, short& count)
{
   //    char *word, *last;
   char *word;
   char *temp = strdup(str);

   for ( word = strtok_r(temp, ";", &last), count = 0;
         word;
         word = strtok_r(NULL, ";", &last), count++)
   {
      intervals[count].type = (SignType) atoi(word);
      if ((word = strchr(word, ',')) == NULL)
         ScThrowErr("Badly formatted intervals list");
      word++;
      intervals[count].start = atof(word);
      if ((word = strchr(word, ',')) == NULL)
         ScThrowErr("Badly formatted intervals list");
      word++;
      intervals[count].end = atof(word);
   }

   free(temp);
}

void TrackView::loadMap(const char *fullFilename)
{
	char filename[1024];
   char *temp;
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullFilename);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandle(xmlDoc);
	
	TiXmlElement *xTrack = docHandle.FirstChildElement("track").Element();
	TiXmlText *xTileset = docHandle.FirstChildElement("track").FirstChild("tileset").FirstChild().ToText();
	TiXmlText *xTilemap = docHandle.FirstChildElement("track").FirstChild("tilemap").FirstChild().ToText();
	TiXmlElement *xStartline = docHandle.FirstChildElement("track").FirstChildElement("startline").Element();
	TiXmlText *xPath = docHandle.FirstChildElement("track").FirstChild("path").FirstChild().ToText();
	TiXmlText *xRoadSigns = docHandle.FirstChildElement("track").FirstChild("roadsigns").FirstChild().ToText();
	
	// do error checking
	if (!(xTrack && xTileset && xTilemap && xStartline && xPath))
	{
		xmlDoc->Clear();
		delete xmlDoc;
		if (!xTrack)
			ScThrowErr("xml track element not found");
		if (!xTileset)
			ScThrowErr("xml tileset element not found");
		if (!xTilemap)
			ScThrowErr("xml tilemap element not found");
		if (!xStartline)
			ScThrowErr("xml startline element not found");
		if (!xPath)
			ScThrowErr("xml path element not found");
		if (!xRoadSigns)
			ScThrowErr("xml roadsigns element not found");
	}
	
   if (xRoadSigns == NULL)
      printf("warning: xml roadsigns element not found. Continuing without it.\n");
   
	// load the tilemap from the specified tilemap file
	strncpy(filename, fullFilename, 1023);
	temp = (char *)strrchr(filename, '/');
	if (!temp)
		ScThrowErr("xml filename is not absolute");
   temp[1] = '\0';
	strcat(filename, xTilemap->Value());
	try { tiles->loadMap(filename); }
	catch (ScException e) {
      e.printMsg();
      newMap(80,50);
      return;
   }
   
	// the path
	pathPolygon.clear();
	if (xPath != NULL) {
		Tuple points[16384];
		short pointsCount;
		readPointsFromString(xPath->Value(), points, pointsCount);
		for (short k = 0; k < pointsCount; k++) {
			pathPolygon.push_back(points[k] * tiles->getTileWidth() / kScale);
		}
	}
   
	// the roadsigns
	signIntervals.clear();
	if (xRoadSigns != NULL) {
		RoadSignInterval intervals[16384];
		short count;
		readIntervalsFromString(xRoadSigns->Value(), intervals, count);
		for (short k = 0; k < count; k++) {
         intervals[k].startPoint = computePathLocation(intervals[k].start);
         intervals[k].endPoint = computePathLocation(intervals[k].end);
			signIntervals.push_back(intervals[k]);
		}
	}
	
	// get the start line attributes
	Tuple p1, p2;
   const char *directionStr;
	if (!(xStartline->Attribute("x1", &p1.x) && xStartline->Attribute("y1", &p1.y) &&
			xStartline->Attribute("x2", &p2.x) && xStartline->Attribute("y2", &p2.y) && (directionStr = xStartline->Attribute("direction"))))
	{
		xmlDoc->Clear();
		delete xmlDoc;
		ScThrowErr("xml startline attributes not set");
	}
	p1.x *= tiles->getTileWidth();
	p2.x *= tiles->getTileWidth();
	p1.y *= tiles->getTileHeight();
	p2.y *= tiles->getTileHeight();
	if (strcmp(directionStr, "clockwise") == 0) {
      clockwise = true;
   }
   else if (strcmp(directionStr, "anti-clockwise") == 0) {
      clockwise = false;
   }
   else {
      ScThrowErr("xml startline attribute direction is invalid");
   }
   
	startLine->setLine(p1, p2);
	if (clockwise)
      ui->clockwiseMenuItem->set();
   else
      ui->clockwiseMenuItem->clear();
   
	// Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
	
	printf("loading done\n");
   redraw();
   ui->setChanged(false);
   resetScroll();
}

void TrackView::init()
{
   try {
	ScTilemap* tilemap = loadTileset("default");	// will get deleted when animation destructor is called
	
	animation.push_back(tilemap);
	tiles = tilemap;
   printf("Current window dims: %d %d\n", this->w(), this->h());
	ScPortRect portRect = {0, 0, this->w(), this->h()};
	ScRect viewRect = {0, 0, portRect.width, portRect.height};
	view = new ScView(portRect, viewRect, animation);
	tilemap->newMap(80, 50, kWallTile);
	
	//list<Player*> emptyPlayersList;
	//minimap = new Minimap(*tilemap, emptyPlayersList);
	
	animation.push_back(new PathLine(tilemap->getBounds(), &pathPolygon));
	
	startLine = new Startline(Tuple(viewRect.right*0.5,viewRect.top*0.75+viewRect.bottom*0.25),
										 Tuple(viewRect.right*0.5,viewRect.top*0.25+viewRect.bottom*0.75),
										 kDefaultStartLineWidth, true);
	animation.push_back(startLine);
   
   char signsFile[1024];
   strcpy(signsFile, dataDir);
   strcat(signsFile, "/images/signs.png");
   //printf("f = %s\n", signsFile);
   Fl_Image * signsImg = loadImage(signsFile);
   if (signsImg == NULL) {
      ScThrowErr("Could not find signs image file");
   }
   
   ScPixelData signsData;
   PixelDataForImage(signsImg, signsData);
   signs = new ScSprite(64, 64, 0, signsData);
   delete signsImg;
   
   updateScrollBars();
   }
   catch (ScException e) {
      printf("Failed to initialise. Make sure we are in the same directory as the toycars app bundle.\n");
      e.printMsg();
      exit(1);
   }
}

double TrackView::computeTotalPathLength() {
   double dist = 0;
   list<Tuple>::const_iterator tmp;
   for (list<Tuple>::const_iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
      tmp = it;
      tmp++;
      if (tmp == pathPolygon.end())
         tmp = pathPolygon.begin();
      dist += (*tmp - *it).Length();
   }
   return dist;
}

/*
double TrackView::computeCursorPathPosition() {
   if (pathPolygon.size() > 1) {
      if (hoverPathVertexExists) {
         double dist = 0;
         list<Tuple>::const_iterator tmp;
         for (list<Tuple>::const_iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
            if (it == hoverPathVertex)
               return dist;
            tmp = it;
            tmp++;
            if (tmp == pathPolygon.end())
               tmp = pathPolygon.begin();
            dist += (*tmp - *it).Length();
         }
      }
      else if (hoverPathIntermediateVertexExists) {
         double dist = 0;
         list<Tuple>::const_iterator tmp, it;
         for (it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
            tmp = it;
            tmp++;
            if (tmp == pathPolygon.end())
               tmp = pathPolygon.begin();
            if (tmp == hoverPathIntermediateNeighbour)
               break;
            dist += (*tmp - *it).Length();
         }
         dist += (*it - hoverPathIntermediateVertex).Length();
         return dist;
      }
   }
   
   return -1;
}
*/

double TrackView::computeNearestPathPosition(Tuple pos) {
   double bestPathDist = 0;
   double curPathDist = 0;
   double bestNormDist = INFINITY;
   Tuple p, norm;
   double len, x, dist;
   for (list<Tuple>::iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
      list<Tuple>::iterator next = it;
      next++;
      if (next == pathPolygon.end())
         next = pathPolygon.begin();
      // project pos onto the line it..next
      p = *next - *it;
      len = p.Normalise();
      x = p * (pos - *it);
      // check if this edge is the closest to pos so far
      if (0 <= x && x <= len) {
         norm = (x * p) + *it;
         dist = (norm - pos).Norm();
         if (dist < bestNormDist) {
            bestNormDist = dist;
            bestPathDist = curPathDist + x;
         }
      }
      // check if this vertrex is the closest to pos so far
      dist = (*it - pos).Norm();
      if (dist < bestNormDist) {
         bestNormDist = dist;
         bestPathDist = curPathDist;
      }
      curPathDist += len;
   }
   return bestPathDist;
}

int TrackView::handle(int event) {
   int x, y;
   Tuple mpos;
   if (view) {
      x = Fl::event_x();
      y = h() - Fl::event_y();
      x += view->getViewRect().left;
      y += view->getViewRect().bottom;
      mpos.x = x / kScale;
      mpos.y = y / kScale;
   }
   switch(event) {
      case FL_FOCUS:
      case FL_UNFOCUS:
         return 1;
      case FL_KEYDOWN:
         if (Fl::event_key() == FL_Delete) {
            if (tool == kRoadSignTool) {
               if (signIntervalStartMouse) {
                  for (list<RoadSignInterval>::iterator it = signIntervals.begin(); it != signIntervals.end(); it++) {
                     if (&(*it) == signIntervalStartMouse) {
                        signIntervals.erase(it);
                        signIntervalStartMouse = NULL;
                        signIntervalEndMouse = NULL;
                        redraw();
                        break;
                     }
                  }
               }
               else if (signIntervalEndMouse) {
                  for (list<RoadSignInterval>::iterator it = signIntervals.begin(); it != signIntervals.end(); it++) {
                     if (&(*it) == signIntervalEndMouse) {
                        signIntervals.erase(it);
                        signIntervalStartMouse = NULL;
                        signIntervalEndMouse = NULL;
                        redraw();
                        break;
                     }
                  }
               }
            }
            else if (selectedPathVertexExists) {
               pathPolygon.erase(selectedPathVertex);
               if (pathPolygon.empty()) {
                  selectedPathVertexExists = false;
               }
               else {
                  selectedPathVertexExists = true;
                  selectedPathVertex = pathPolygon.end();
                  selectedPathVertex--;
               }
               hoverPathVertexExists = false;
               ui->setChanged(true);
               redraw();
               return 1;
            }
         }
         return Fl_Gl_Window::handle(event);
      case FL_ENTER:
      case FL_LEAVE:
         return 1;
      case FL_MOVE:
      {
            bool oldHoverPathIntermediateVertexExists = hoverPathIntermediateVertexExists;
            bool oldHoverPathVertexExists = hoverPathVertexExists;
            hoverPathIntermediateVertexExists = false;
            hoverPathVertexExists = false;
            if (tool == kRoadSignTool) {
               signIntervalEndMouse = NULL;
               signIntervalStartMouse = NULL;
               signEnd = computeNearestPathPosition(mpos);
               if (signEnd < signStart)
                  signEnd += computeTotalPathLength();
               if (drawingSignInterval)
                  redraw();
               for (list<RoadSignInterval>::iterator it = signIntervals.begin(); it != signIntervals.end(); it++) {
                  if (sqr(mpos.x - it->startPoint.x) + sqr(mpos.y - it->startPoint.y) < sqr(kIntervalHandleRadius)) {
                     signIntervalStartMouse = &(*it);
                  }
                  if (sqr(mpos.x - it->endPoint.x) + sqr(mpos.y - it->endPoint.y) < sqr(kIntervalHandleRadius)) {
                     signIntervalEndMouse = &(*it);
                  }
               }
            }
            if (tool == kPathTool || tool == kRoadSignTool) {
               
               // look to see if hovering cursor over a vertex in the path polygon
               for (list<Tuple>::iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
                  if (((*it) - mpos).Norm() < sqr(kVertexSelectMinDist)) {
                     hoverPathVertex = it;
                     hoverPathVertexExists = true;
                  }
               }
               if (oldHoverPathVertexExists != hoverPathVertexExists)
                  redraw();
               if (hoverPathVertexExists) {
                  return 1;
               }
               
               // look to see if hovering cursor over a point on an edge of the path polygon
               if (pathPolygon.size() <= 1)
                  return 1;
               for (list<Tuple>::iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
                  list<Tuple>::iterator next = it;
                  next++;
                  if (next == pathPolygon.end())
                     next = pathPolygon.begin();
                  // project mpos onto the line it..next
                  Tuple p = *next - *it;
                  double len = p.Normalise();
                  double x = p * (mpos - *it);
                  if (0 <= x && x <= len) {
                     hoverPathIntermediateVertex = (x * p) + *it;
                     if ((hoverPathIntermediateVertex - mpos).Norm() < sqr(kPathSelectMinDist)) {
                        hoverPathIntermediateVertexExists = true;
                        hoverPathIntermediateNeighbour = next;
                        redraw();
                        break;
                     }
                  }
               }
               if (oldHoverPathIntermediateVertexExists != hoverPathIntermediateVertexExists)
                  redraw();
            }
         }
         return 1;
      case FL_DRAG:
      case FL_PUSH: {
         int b = Fl::event_button();
         if (event == FL_PUSH) {
            click_x = x;
            click_y = y;
            drag_x = Fl::event_x();
            drag_y = Fl::event_y();
         }
         const ToolType saveTool = tool;
         if (b == FL_RIGHT_MOUSE)
            tool = kScrollTool;
         switch (tool) {
            case kScrollTool:
               if (event == FL_DRAG) {
                  view->scrollBy(drag_x - Fl::event_x(), Fl::event_y() - drag_y);
                  boundScroll();
                  updateScrollBars();
               }
               break;
            case kDrawTool:
               if (b == FL_LEFT_MOUSE || b == FL_RIGHT_MOUSE) {
                  int col = x / tiles->getTileWidth();
                  int row = y / tiles->getTileHeight();
                  int t=0;
                  switch (tile) {case kRoad: t=(col%5)+5*(row%5)+kStartRoadTiles; break; case kOffroad: t=(col%5)+5*(row%5)+kStartOffroadTiles; break; case kWall: t=kWallTile; break;}
                  tiles->getTileAtPoint(x,y) = b == FL_LEFT_MOUSE ? t : kWallTile;
                  ui->setChanged(true);
               }
               break;
            case kPathTool:
               if (hoverPathVertexExists && b == FL_LEFT_MOUSE) {
                  ui->setChanged(true);
                  selectedPathVertex = hoverPathVertex;
                  if (event == FL_DRAG)
                     *selectedPathVertex = mpos;
                  selectedPathVertexExists = true;
                  ui->setChanged(true);
                  break;
               }
               if (hoverPathIntermediateVertexExists && b == FL_LEFT_MOUSE) {
                  ui->setChanged(true);
                  pathPolygon.insert(hoverPathIntermediateNeighbour, hoverPathIntermediateVertex);
                  list<Tuple>::iterator temp = hoverPathIntermediateNeighbour;
                  temp--;
                  selectedPathVertex = temp;
                  hoverPathVertex = selectedPathVertex;
                  hoverPathIntermediateVertexExists = false;
                  hoverPathVertexExists = true;
                  selectedPathVertexExists = true;
                  ui->setChanged(true);
                  break;
               }
               if (b == FL_LEFT_MOUSE) {
                  ui->setChanged(true);
                  if (selectedPathVertexExists && Fl::event_shift()) {
                     selectedPathVertex->x = x/kScale;
                     selectedPathVertex->y = y/kScale;
                  }
                  else {
                     pathPolygon.push_back(Tuple(x,y)/kScale);
                     hoverPathVertex = pathPolygon.end();
                     hoverPathVertex--;
                     selectedPathVertex = hoverPathVertex;
                     hoverPathVertexExists = true;
                     selectedPathVertexExists = true;
                  }
               }
               break;
            case kStartlineTool:
               if (b == FL_LEFT_MOUSE) {
                  ui->setChanged(true);
                  startLine->setLine(Tuple(click_x,click_y),Tuple(x,y));
               }
               if (b == FL_RIGHT_MOUSE) {
                  ui->setChanged(true);
                  startLine->setLine(startLine->getStartPoint(),Tuple(x,y));
               }
               break;
            case kRoadSignTool:
               if (b == FL_LEFT_MOUSE) {
                  ui->setChanged(true);
                  if (signIntervalStartMouse) {
                     signIntervalStartMouse->start = computeNearestPathPosition(mpos);
                     signIntervalStartMouse->startPoint = computePathLocation(signIntervalStartMouse->start);
                     const double pathLen = computeTotalPathLength();
                     if (signIntervalStartMouse->start > signIntervalStartMouse->end)
                        signIntervalStartMouse->end += pathLen;
                     else if (signIntervalStartMouse->end - signIntervalStartMouse->start > pathLen)
                        signIntervalStartMouse->end -= pathLen;
                  }
                  else if (signIntervalEndMouse) {
                     signIntervalEndMouse->end = computeNearestPathPosition(mpos);
                     signIntervalEndMouse->endPoint = computePathLocation(signIntervalEndMouse->end);
                     const double pathLen = computeTotalPathLength();
                     if (signIntervalEndMouse->start > signIntervalEndMouse->end)
                        signIntervalEndMouse->end += pathLen;
                     else if (signIntervalEndMouse->end - signIntervalEndMouse->start > pathLen)
                        signIntervalEndMouse->end -= pathLen;
                  }
                  else if (event == FL_PUSH) {
                     //double curPos = computeCursorPathPosition();
                     double curPos = computeNearestPathPosition(mpos);
                     if (drawingSignInterval) {
                        drawingSignInterval = false;
                        if (curPos < signStart)
                           curPos += computeTotalPathLength();
                        signIntervals.push_back(RoadSignInterval(sign, signStart, curPos,
                                                computePathLocation(signStart), computePathLocation(curPos)));
                     }
                     else {
                        drawingSignInterval = true;
                        signStart = curPos;
                        signEnd = signStart;
                     }
                  }
               }
               break;
         }
         tool = saveTool;
         if (event == FL_DRAG) {
            drag_x = Fl::event_x();
            drag_y = Fl::event_y();
         }
         redraw();
         return 1;
      }/*
      case FL_MOUSEWHEEL:
         view->scrollBy(Fl::event_dx(), -Fl::event_dy());
         boundScroll();
         updateScrollBars();
         redraw();
         return 1;*/
      /*
      case FL_RELEASE:
         printf("b: %d x: %d y: %d\n", Fl::event_button(), Fl::event_x(), Fl::event_y());
         redraw();
         return 1;
      case FL_SHORTCUT:
         if (Fl::event_key() == 'x') {
            printf("you pressed x\n");
         }
         redraw();
         return 0;
      */
      default:
         return Fl_Gl_Window::handle(event);
   }
}

void TrackView::resetImportImageDialog(const char *filename)
{
   Fl_Image *img = loadImage(filename);
   if (img == NULL)
      return;
   ui->importImageBox->image(img);
   ui->importImageBox->setFilename(filename);
   ui->roadColourButton->color(FL_BACKGROUND_COLOR);
   ui->offroadColourButton->color(FL_BACKGROUND_COLOR);
   ui->wallColourButton->color(FL_BACKGROUND_COLOR);
   ui->roadColourButton->clear();
   ui->offroadColourButton->clear();
   ui->wallColourButton->clear();
   ui->roadColourButton->set();
}

bool compareColours(const unsigned char *c1, const unsigned char *c2)
{
   return c1[0] == c2[0] && c1[1] == c2[1] && c1[2] == c2[2];
}

void TrackView::importImage(const char *fullFilename)
{
	Fl_Image	*img;
	unsigned short width, height, bpp, pad_bytes;
	unsigned char value8;
	int i, j;
	unsigned char *cur;
	
	/* Try to load the image file */
	img = loadImage(fullFilename);
	if (img == NULL) {
		ScThrowErr("Error opening image file");
	}
	
	/* calculate some numbers */
	width = img->w();
	height = img->h();
	bpp = img->d();
	/*pad_bytes = (4 - (width%4)) % 4;*/
	//pad_bytes = surface->pitch/bpp - width;
	pad_bytes = 0;
   
	//printf("testing: %d should be equal to %d\n", pad_bytes, (4 - (width%4)) % 4);
	
   printf("bpp: %d count: %d\n", bpp, img->count());
	
	newMap(width, height);
   
	/* Write map tile data to tilemap */
	cur = (unsigned char *)img->data()[0];
	for (i = 0; i < height; i++)
	{
		for (j = 0; j < width; j++)
		{
         if (compareColours(cur, ui->importImageBox->getRoadColour()))
            value8 = 5*(i%5) + (j%5) + kStartRoadTiles;
         else if (compareColours(cur, ui->importImageBox->getOffroadColour()))
            value8 = 5*(i%5) + (j%5) + kStartOffroadTiles;
         else// if (compareColours(cur, importImageBox->getWallColour())
            value8 = kWallTile;
         
			(*tiles)(j,height-i-1) = value8;
			cur += bpp;
		}
		cur += pad_bytes;
	}
	
	delete img;
	
	printf("loading from image done\n");
   ui->setChanged(false);
   resetScroll();
}

// installs map into user data directory
void TrackView::installMap(const char *name)
{
#ifdef WIN32
   fl_alert("Unfortunately this is unsupported on Windows at this time.\nYou may still install maps manually by saving them to a folder and placing it inside the data\\tracks directory and editing the tracklist.xml file.");
   return;
#else
   char fullname[1024];
   char trackname[256];
   const char kTracklistName[64] = "/tracks/tracklist.xml";
   char *temp;
   
   // first we uninstall the named track
   uninstallMap(name);
   
   // chop off any .xml suffix
   temp = strrchr(name, '.');
   if (temp) {
      if (strcmp(temp, ".xml") == 0)
         temp[0] = '\0';
   }
   // remove path
   temp = strrchr(name, '/');
   if (temp) {
      strncpy(trackname, temp+1, 255);
   }
   
   strncpy(fullname, userDataDir, 1023 - strlen(kTracklistName));
   strcat(fullname, kTracklistName);
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullname);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	// Add name to track list
	TiXmlHandle *docHandle = new TiXmlHandle(xmlDoc);
	TiXmlElement *xTracklist = docHandle->FirstChildElement("tracklist").Element();
	if (xTracklist == NULL)
		ScThrowErr("xml file could not find tracklist element");
   TiXmlElement track("track");
   track.SetAttribute("name", trackname);
   xTracklist->InsertEndChild(track);
	
	// Save the xml file
	if (!xmlDoc->SaveFile())
	{
		delete xmlDoc;
		ScThrowErr("error saving xml file");
	}
	
	//delete xTracklist;
	delete docHandle;
	delete xmlDoc;
	
   strncpy(fullname, userDataDir, 1023 - 8);
   strcat(fullname, "/tracks/");
   strcat(fullname, trackname);

   // make sure the tracks directory exists. nb: what whould be done on win32 here???
   char str[1024];
   sprintf(str, "mkdir -p %s", fullname);
   system(str);

   strcat(fullname, "/");
   strcat(fullname, trackname);
   strcat(fullname, ".xml");
   saveMap(fullname);
   
	printf("install done\n");
#endif
}

// uninstall track from user data directory.
// nb: doesn't remove track data itself, only remove entry from tracklist xml file.
void TrackView::uninstallMap(const char *name)
{
#ifdef WIN32
   fl_alert("Unfortunately this is unsupported on Windows at this time.\nYou may still uninstall maps manually by removing them from the data\\tracks directory and editing the tracklist.xml file.");
   return;
#else
   char fullname[1024];
   char trackname[256];
   const char kTracklistName[64] = "/tracks/tracklist.xml";
	const char *attribute;
   char *temp;
   
   // chop off any .xml suffix
   temp = strrchr(name, '.');
   if (temp) {
      if (strcmp(temp, ".xml") == 0)
         temp[0] = '\0';
   }
   // remove path
   temp = strrchr(name, '/');
   if (temp) {
      strncpy(trackname, temp+1, 255);
   }
   
   strncpy(fullname, userDataDir, 1023 - strlen(kTracklistName));
   strcat(fullname, kTracklistName);
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullname);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	// Remove all occurrences of the track the from track list
	TiXmlHandle *docHandle = new TiXmlHandle(xmlDoc);
	TiXmlElement *xTracklist = docHandle->FirstChildElement("tracklist").Element();
	if (xTracklist == NULL)
		ScThrowErr("xml file could not find tracklist element");
   TiXmlElement *track = xTracklist->FirstChildElement("track");
   TiXmlElement *prevTrack;
	while (track != NULL)
	{
		attribute = track->Attribute("name");
		if (attribute == NULL)
			ScThrowErr("xml track element does not have attribute called \"name\"");
		prevTrack = track;
		track = track->NextSiblingElement("track");
      if (strcmp(attribute, trackname) == 0) {
         xTracklist->RemoveChild(prevTrack);
      }
	}
   
   // Save the xml file
   if (!xmlDoc->SaveFile()) {
      delete xmlDoc;
      ScThrowErr("error saving xml file");
   }
	
	//delete xTracklist;
	delete docHandle;
	delete xmlDoc;
   
	printf("uninstall done\n");
#endif
}

void TrackView::saveMap(const char *fullFilename)
{
	char filename[1024] = "";
   char trackname[256] = "";
   char tempStr[256];
   char longStr[16384] = "";
   char *suffix;
   Tuple startLineA = startLine->getStartPoint();
   Tuple startLineB = startLine->getEndPoint();
   
	// Setup the xml file
	//printf("xml filename is %s\n", filename);
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullFilename);
	
   TiXmlDeclaration decl("1.0", "UTF-8", "yes");
	TiXmlElement xTrack("track");
	TiXmlElement xTileset("tileset");
	TiXmlElement xTilemap("tilemap");
	TiXmlElement xStartline("startline");
	TiXmlElement xPath("path");
	TiXmlElement xRoadSigns("roadsigns");
	
	xTileset.InsertEndChild(TiXmlText("default"));
	
	startLineA.x /= tiles->getTileWidth();
	startLineA.y /= tiles->getTileHeight();
	startLineB.x /= tiles->getTileWidth();
	startLineB.y /= tiles->getTileHeight();
	xStartline.SetDoubleAttribute("x1", startLineA.x);
	xStartline.SetDoubleAttribute("y1", startLineA.y);
	xStartline.SetDoubleAttribute("x2", startLineB.x);
	xStartline.SetDoubleAttribute("y2", startLineB.y);
	xStartline.SetAttribute("direction", clockwise ? "clockwise" : "anti-clockwise");
	
   strncpy(filename, fullFilename, 1023);
   suffix = strrchr(filename, '.');
   if (!suffix)
      ScThrowErr("File has no suffix");
   if (strcmp(suffix,".xml") != 0)
      ScThrowErr("File suffix is not .xml");
   suffix[0] = '\0';
   
   suffix = strrchr(filename, '/');
   suffix[0] = '\0';
   if (!suffix)
      ScThrowErr("File path is not absolute");
   strncpy(trackname, &suffix[1], 255);
   
	strcat(trackname, ".map");
	xTilemap.InsertEndChild(TiXmlText(trackname));
	
   // create the path element
	for (list<Tuple>::const_iterator i = pathPolygon.begin(); i != pathPolygon.end(); i++) {
		sprintf(tempStr, "%.2lf,%.2lf; ", i->x / tiles->getTileWidth() * kScale, i->y / tiles->getTileWidth() * kScale);
		strcat(longStr, tempStr);
	}
	if (!pathPolygon.empty()) {
		longStr[strlen(longStr) - 2] = '\0';
		xPath.InsertEndChild(TiXmlText(longStr));
	}
	else {
		xPath.InsertEndChild(TiXmlText(""));
	}
   
   // create the roadsigns element
   longStr[0] = '\0';
	for (list<RoadSignInterval>::const_iterator i = signIntervals.begin(); i != signIntervals.end(); i++) {
		sprintf(tempStr, "%d,%.2lf,%.2lf; ", i->type, i->start, i->end);
		strcat(longStr, tempStr);
	}
	if (!signIntervals.empty()) {
		longStr[strlen(longStr) - 2] = '\0';
		xRoadSigns.InsertEndChild(TiXmlText(longStr));
	}
	else {
		xRoadSigns.InsertEndChild(TiXmlText(""));
	}
	
	xTrack.InsertEndChild(xTileset);
	xTrack.InsertEndChild(xTilemap);
	xTrack.InsertEndChild(xStartline);
	xTrack.InsertEndChild(xPath);
	xTrack.InsertEndChild(xRoadSigns);
	
	strcat(filename, "/");
	strcat(filename, trackname);
   xmlDoc->InsertEndChild(decl);
   xmlDoc->InsertEndChild(xTrack);
	xmlDoc->SaveFile();
	printf("Saving to file %s...\n", filename);
	tiles->saveMap(filename);
	
	// Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
	
	printf("saving done\n");
   ui->setChanged(false);
}

void TrackView::updateScrollBars() {
   const float viewHeight = view->getViewRect().top-view->getViewRect().bottom;
   const float mapHeight = tiles->getBounds().top-tiles->getBounds().bottom;
   const float scrollCentreV = 0.5*(view->getViewRect().bottom+view->getViewRect().top);
   const float scrollBarHeight = viewHeight/mapHeight;
   
   const float viewWidth = view->getViewRect().right-view->getViewRect().left;
   const float mapWidth = tiles->getBounds().right-tiles->getBounds().left;
   const float scrollCentreH = 0.5*(view->getViewRect().right+view->getViewRect().left);
   const float scrollBarWidth = viewWidth/mapWidth;
   
   ui->vertScroll->value((1 - (scrollCentreV - 0.5*viewHeight)/(mapHeight - viewHeight))*(1-scrollBarHeight)*kScrollbarMaxValue, scrollBarHeight*kScrollbarMaxValue, 0, kScrollbarMaxValue);
   ui->horizScroll->value(((scrollCentreH - 0.5*viewWidth)/(mapWidth - viewWidth))*(1-scrollBarWidth)*kScrollbarMaxValue, scrollBarWidth*kScrollbarMaxValue, 0, kScrollbarMaxValue);
}

void TrackView::drawTrack() {
   view->draw();
}

void TrackView::resize(int x, int y, int w, int h)
{
   Fl_Gl_Window::resize(x,y,w,h);
   view->resize(this->w(), this->h());
   boundScroll();
   updateScrollBars();
   redraw();
}

void TrackView::setHorizScrollFrac(double x)
{
   const float width = view->getViewRect().right - view->getViewRect().left;
   view->scrollTo(floor(x*tiles->getBounds().right + 0.5*width), floor(0.5*(view->getViewRect().bottom+view->getViewRect().top)));
   boundScroll();
   redraw();
}

void TrackView::setVertScrollFrac(double x)
{
   const float height = view->getViewRect().top - view->getViewRect().bottom;
   view->scrollTo(floor(0.5*(view->getViewRect().left+view->getViewRect().right)), floor(tiles->getBounds().top - (x*tiles->getBounds().top + 0.5*height)));
   boundScroll();
   redraw();
}

void TrackView::boundScroll()
{
   if (view->getViewRect().left < tiles->getBounds().left)
      view->scrollBy(-view->getViewRect().left, 0);
   if (view->getViewRect().bottom < tiles->getBounds().bottom)
      view->scrollBy(0, -view->getViewRect().bottom);
   if (view->getViewRect().right > tiles->getBounds().right)
      view->scrollBy(tiles->getWidth() - view->getViewRect().right, 0);
   if (view->getViewRect().top > tiles->getBounds().top)
      view->scrollBy(0, tiles->getHeight() - view->getViewRect().top);
}

Tuple TrackView::computePathLocation(const double loc) {
   if (pathPolygon.size() == 1)
      return pathPolygon.front();
   if (pathPolygon.empty())
      ScThrowErr("Attempt to compute path location but no path exists.");
   /*
   if (loc < 0)
      printf("warning: negative path loc\n");//ScThrowErr("Attempt to compute path location with negative value.");
   */
   double dist = 0;
   list<Tuple>::const_iterator tmp;
   while (1) {
      for (list<Tuple>::const_iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
         tmp = it;
         tmp++;
         if (tmp == pathPolygon.end())
            tmp = pathPolygon.begin();
         double segmentDist = (*tmp - *it).Length();
         if (dist + segmentDist >= loc) {
            double offset = loc - dist;
            Tuple N = (*tmp - *it).Direction() * offset;
            return *it + N;
         }
         dist += segmentDist;
      }
   }
   
   return Tuple(0,0);
}

void TrackView::drawSignInterval(RoadSignInterval &interval, bool drawSign) {
   const Tuple A = computePathLocation(interval.start);
   const Tuple B = computePathLocation(interval.end);
   interval.startPoint = A;
   interval.endPoint = B;
   double dist = 0;
   list<Tuple>::const_iterator tmp;
   bool pastStart = false;
   bool drawingPath = true;
   
   // draw the path interval
   glLineWidth(12);
   glColor4f(1,1,1,0.5);
   glDisable(GL_TEXTURE_2D);
   glEnable(GL_TEXTURE_1D);
   glBegin(GL_LINES);
   
   while (drawingPath) {
      for (list<Tuple>::const_iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
         tmp = it;
         tmp++;
         if (tmp == pathPolygon.end())
            tmp = pathPolygon.begin();
         dist += (*tmp - *it).Length();
         if (dist > interval.start) {
            if (!pastStart) {
               if (dist > interval.end) {
                  glVertex2d(A.x * kScale, A.y * kScale);
                  glVertex2d(B.x * kScale, B.y * kScale);
                  drawingPath = false;
                  break;
               }
               pastStart = true;
               // draw from A to *it
               glVertex2d(A.x * kScale, A.y * kScale);
               glVertex2d(tmp->x * kScale, tmp->y * kScale);
               continue;
            }
         }
         if (dist > interval.end && pastStart) {
            // draw from *it to B
            glVertex2d(it->x * kScale, it->y * kScale);
            glVertex2d(B.x * kScale, B.y * kScale);
            drawingPath = false;
            break;
         }
         // draw from *it to *tmp
         if (pastStart) {
            glVertex2d(it->x * kScale, it->y * kScale);
            glVertex2d(tmp->x * kScale, tmp->y * kScale);
         }
      }
   }
   
   glEnd();
   
   // draw the actual sign
   if (drawSign) {
      glEnable(GL_TEXTURE_2D);
      glDisable(GL_TEXTURE_1D);
      signs->setFrame(interval.type);
      signs->moveTo(A.x * kScale, A.y * kScale);
      signs->doDraw();
   }
   
   // draw handles
   glDisable(GL_TEXTURE_2D);
   glEnable(GL_TEXTURE_1D);
   glColor4f(0.2, 0, 1, 1);
   drawCircleAt(A.x, A.y, kIntervalHandleRadius);
   glColor4f(1, 0, 0.2, 1);
   drawCircleAt(B.x, B.y, kIntervalHandleRadius);
}

void TrackView::draw() {
   if (!view)
      init();
   if (!valid()) {
      //glLoadIdentity();
      //glViewport(0,0,w(),h());
      //glOrtho(0, w(), 0, h(), -1, 1);
      //ortho();
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   }
   
   glClear(GL_COLOR_BUFFER_BIT);
   
   glPushMatrix();
   
   drawTrack();
   
   // draw path and road signs
   if (hoverPathVertexExists) {
      if (tool == kRoadSignTool) {
         if (!(signIntervalStartMouse || signIntervalEndMouse || drawingSignInterval)) {
            signs->setFrame(sign);
            signs->moveTo(hoverPathVertex->x * kScale, hoverPathVertex->y * kScale);
            signs->doDraw();
         }
      }
      else {
         glColor4f(1.0, 1.0, 1.0, 0.5);
         glDisable(GL_TEXTURE_2D);
         glEnable(GL_TEXTURE_1D);
         drawCircleAt(hoverPathVertex->x, hoverPathVertex->y, kVertexSelectMinDist);
      }
   }
   if (hoverPathIntermediateVertexExists) {
      if (tool == kRoadSignTool) {
         if (!(signIntervalStartMouse || signIntervalEndMouse || drawingSignInterval)) {
            signs->setFrame(sign);
            signs->moveTo(hoverPathIntermediateVertex.x * kScale, hoverPathIntermediateVertex.y * kScale);
            signs->doDraw();
         }
      }
      else {
         glColor4f(0.0, 1.0, 1.0, 0.5);
         glDisable(GL_TEXTURE_2D);
         glEnable(GL_TEXTURE_1D);
         drawCircleAt(hoverPathIntermediateVertex.x, hoverPathIntermediateVertex.y, kPathSelectMinDist);
      }
   }
   if (selectedPathVertexExists) {
      glLineWidth(1);
      glColor4f(1.0, 0, 0, 0.7);
      glDisable(GL_TEXTURE_2D);
      glEnable(GL_TEXTURE_1D);
      drawCircleAt(selectedPathVertex->x, selectedPathVertex->y, kVertexSelectMinDist);
      glColor4f(1.0, 1.0, 1.0, 1.0);
      drawCircleAt(selectedPathVertex->x, selectedPathVertex->y, kVertexSelectMinDist, false);
   }
   
   // draw road sign intervals
   if (tool == kRoadSignTool || tool == kPathTool) {
      for (list<RoadSignInterval>::iterator it = signIntervals.begin(); it != signIntervals.end(); it++) {
         drawSignInterval(*it, tool == kRoadSignTool);
      }
   }
   
   // drawing interval
   if (drawingSignInterval) {
      RoadSignInterval i(sign, signStart, signEnd, computePathLocation(signStart), computePathLocation(signEnd));
      drawSignInterval(i, false);
   }
   
   // draw selection of interval handle
   glLineWidth(1);
   glColor4f(1, 1, 1, 1);
   glDisable(GL_TEXTURE_2D);
   glEnable(GL_TEXTURE_1D);
   if (signIntervalStartMouse) {
      drawCircleAt(signIntervalStartMouse->startPoint.x, signIntervalStartMouse->startPoint.y, kIntervalHandleRadius, false);
   }
   if (signIntervalEndMouse) {
      drawCircleAt(signIntervalEndMouse->endPoint.x, signIntervalEndMouse->endPoint.y, kIntervalHandleRadius, false);
   }
   
   glPopMatrix();
}

void swap(unsigned char &x, unsigned char &y) {
   unsigned char z = x;
   x = y;
   y = z;
}

void TrackView::flipHorizontal() {
   const short h = tiles->getMapHeight();
   const short w = tiles->getMapWidth();
   const short W = tiles->getMapWidth() * tiles->getTileWidth();
   Tuple p1, p2;
   double lineWidth;
   int i, j;
   
   for (i = 0; i < w/2; i++) {
      for (j = 0; j < h; j++) {
         swap((*tiles)(i,j), (*tiles)(w-i-1,j));
      }
   }
   
   for (list<Tuple>::iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
      it->x = W/kScale - it->x;
   }
   pathPolygon.reverse();
   
   p1 = startLine->getStartPoint();
   p2 = startLine->getEndPoint();
   lineWidth = startLine->getWidth();
   
   p1.x = W - p1.x;
   p2.x = W - p2.x;
   
   startLine->setLine(p2,p1,lineWidth);
}

void TrackView::flipVertical() {
   const short h = tiles->getMapHeight();
   const short w = tiles->getMapWidth();
   const short H = tiles->getMapHeight() * tiles->getTileHeight();
   Tuple p1, p2;
   double lineWidth;
   int i, j;
   
   for (i = 0; i < w; i++) {
      for (j = 0; j < h/2; j++) {
         swap((*tiles)(i,j), (*tiles)(i,h-j-1));
      }
   }
   
   for (list<Tuple>::iterator it = pathPolygon.begin(); it != pathPolygon.end(); it++) {
      it->y = H/kScale - it->y;
   }
   pathPolygon.reverse();
   
   p1 = startLine->getStartPoint();
   p2 = startLine->getEndPoint();
   lineWidth = startLine->getWidth();
   
   p1.y = H - p1.y;
   p2.y = H - p2.y;
   
   startLine->setLine(p2,p1,lineWidth);
}
