/*
 *  CurveView.cpp
 *  toycars_vehicle_editor
 *
 *  Created by Ruben Henner Zilibowitz on 23/02/09.
 *  Copyright 2009 Ruben Henner Zilibowitz. All rights reserved.
 *
 */

#include "CurveView.h"
#include "FL/Fl_Image.H"
#include "FL/fl_draw.H"
#include <cmath>
#include <cstdio>

CurveView::CurveView(int x, int y, int w, int h, const char *c) : Fl_Box(x,y,w,h,c), xs(0), oldX(0), oldY(0)
{
   reset();
}

CurveView::CurveView(Fl_Boxtype b, int x, int y, int w, int h, const char *c) : Fl_Box(b,x,y,w,h,c)
{
}

CurveView::~CurveView()
{
}

void CurveView::reset()
{
   x1 = -14;
   x2 = 14;
   y1 = -14;
   y2 = 14;
   gradations = 10;
}

int CurveView::handle(int event)
{
   if (event == FL_PUSH) {
      oldX = Fl::event_x();
      oldY = Fl::event_y();
      return 1;
   }
   if (event == FL_DRAG) {
      double newX = Fl::event_x();
      double newY = Fl::event_y();
      double dx = newX - oldX;
      double dy = newY - oldY;
      x1 -= dx*0.01;
      x2 += dx*0.01;
      y1 -= dy*0.01;
      y2 += dy*0.01;
      oldX = newX;
      oldY = newY;
      redraw();
      return 1;
   }
   if (event == FL_RELEASE) {
      return 1;
   }
   return Fl_Box::handle(event);
}

void CurveView::draw()
{
   const double scaley = 1.0 / (y2 - y1);
   const double scalex = 1.0 / (x2 - x1);
   double dx, xg, yg;
   char str[16];
   
   // clear
   fl_draw_box(FL_FLAT_BOX,x(),y(),w(),h(),FL_WHITE);
   
   // draw axes
   fl_color(FL_GRAY);
   fl_line(x(),y() + h()*0.5,x() + w(),y() + h()*0.5);
   fl_line(x() + w()*0.5,y(),x() + w()*0.5,y() + h());
   fl_font(FL_TIMES, 12);
   dx = (y2-y1) / gradations;
   for (double i = y1; i < y2; i += dx) {
      if (fabs(i) < 0.01)
         continue;
      sprintf(str, "%.1f", i);
      xg = x() + w()*0.5;
      yg = y() + h() - (i*scaley + 0.5)*h();
      fl_line(xg-2,yg,xg+2,yg);
      fl_draw(str, xg + 6, yg + fl_height()/3);
   }
   dx = (x2-x1) / gradations;
   for (double i = x1; i < x2; i += dx) {
      if (fabs(i) < 0.01)
         continue;
      sprintf(str, "%.1f", i);
      xg = (i - x1)*w()*scalex + x();
      yg = y() + 0.5*h();
      fl_line(xg,yg-2,xg,yg+2);
      fl_draw(str, xg - fl_width(str)*0.5, yg + 12);
   }
   fl_draw("0", x() + w()*0.5 - 2*fl_width("0"), y() + 0.5*h() + 12);
   
   // draw graph
   fl_color(FL_BLACK);
   int nx,ny,mx,my;
   dx = 0.01;
   for (double a = x1; a < x2; a += dx) {
      nx = (a - x1)*w()*scalex + x();
      ny = y() + h() - (pacejkaFn(a)*scaley + 0.5)*h();
      mx = (a+dx - x1)*w()*scalex + x();
      my = y() + h() - (pacejkaFn(a+dx)*scaley + 0.5)*h();
      fl_line(nx,ny,mx,my);
   }
}

double CurveView::pacejkaFn(double value)
{
	value *= xs[0];
	return xs[2]*sin(xs[1]*atan(value - xs[3]*(value - atan(value))));
}

// Pacejka magic formula
// D*sin(C*atan(B*theta - E*(B*theta - atan(B*theta))))
// slope at origin == B*C*D
// limit at infinity ==	D*sin(C*pi/2) if E > -1
//						D*sin(C*atan(pi/2)) if E == -1
//						-D*sin(C*pi/2) if E < -1
