/*
 *  VehicleObject.cpp
 *  toycars_vehicle_editor
 *
 *  Created by Ruben Henner Zilibowitz on 23/02/09.
 *  Copyright 2009 Ruben Henner Zilibowitz. All rights reserved.
 *
 */

#include "VehicleObject.h"
#include "ScException.h"
#include "tinyxml.h"
#include "SearchDirectories.h"

#ifndef strtok_r
   #define strtok_r(s,d,p) strtok(s,d)
#endif

extern SearchDirectories *gSearchDirs;

VehicleObject::VehicleObject()
{
   reset();
}

VehicleObject::~VehicleObject()
{
}

void VehicleObject::reset()
{
   title[0] = '\0';
   author[0] = '\0';
   vehicleFrameWidth = 0;
   vehicleFrameHeight = 0;
   vehicleSpriteFilename[0] = '\0';
   geometry.clear();
   axelWidth = 0;
   frontAxel = 0;
   rearAxel = 0;
   frontWheelInset = 0;
   frontWheelRadius = 0;
   rearWheelRadius = 0;
   mass = 0;
   inertia = 0;
   frontWheelInertia = 0;
   rearWheelInertia = 0;
   drag = 0;
   rollingResistance = 0;
   wheelFrameWidth = 0;
   wheelFrameHeight = 0;
   wheelsAboveSprite = 0;
   wheelSpriteFilename[0] = '\0';
   for (int i = 0; i < 4; i++) {
      frontSlipRatioPacejka[i] = 0;
      frontSlipAnglePacejka[i] = 0;
      rearSlipRatioPacejka[i] = 0;
      rearSlipAnglePacejka[i] = 0;
   }
}

// Read the list of points for a convex object. Source taken from main.cpp in toycars.
void readPointsFromString(const char* str, std::list<Vec2D> &points, short& count)
{
    char *word;
    //char *last;
    char *temp = strdup(str);
    Vec2D p;
    
    for (   word = strtok_r(temp, ";", &last), count = 0;
            word;
            word = strtok_r(NULL, ";", &last), count++)
    {
        p.x = atof(word);
        if ((word = strchr(word, ',')) == NULL)
            ScThrowErr("Badly formatted points list");
        p.y = atof(word + 1);
        points.push_back(Vec2D(p.x,p.y));
    }
    
    free(temp);
}

// Read the list of data. Source taken from main.cpp in toycars.
void readDataFromString(const char* str, double *points, short maxcount)
{
    char *word;
    //char *last;
    char *temp = strdup(str);
    short count;
    Vec2D p;
    
    for (   word = strtok_r(temp, ",", &last), count = 0;
            word && count < maxcount;
            word = strtok_r(NULL, ",", &last), count++)
    {
        points[count] = atof(word);
    }
    
    free(temp);
}

void VehicleObject::setTitle(const char *str)
{
   strcpy(title, str);
}

void VehicleObject::setAuthor(const char *str)
{
   strcpy(author, str);
}

void VehicleObject::loadVehicle(const char *fullFilename)
{
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullFilename);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandle(xmlDoc);
	
	TiXmlElement *xCar = docHandle.FirstChildElement("car").Element();
	TiXmlText *xTitle = docHandle.FirstChildElement("car").FirstChild("title").FirstChild().ToText();
	TiXmlText *xAuthor = docHandle.FirstChildElement("car").FirstChild("author").FirstChild().ToText();
	TiXmlElement *xSprite = docHandle.FirstChildElement("car").FirstChildElement("sprite").Element();
	TiXmlText *xSpriteFilename = xSprite->FirstChild()->ToText();
	TiXmlElement *xGeometry = docHandle.FirstChildElement("car").FirstChildElement("geometry").Element();
	TiXmlElement *xProperties = docHandle.FirstChildElement("car").FirstChildElement("properties").Element();
	TiXmlElement *xWheelsprite = docHandle.FirstChildElement("car").FirstChildElement("wheelsprite").Element();
	TiXmlText *xWheelSpriteFilename = xWheelsprite->FirstChild()->ToText();
	TiXmlText *xFrontSR = docHandle.FirstChildElement("car").FirstChild("front_SR_Pacejka").FirstChild().ToText();
	TiXmlText *xFrontSA = docHandle.FirstChildElement("car").FirstChild("front_SA_Pacejka").FirstChild().ToText();
	TiXmlText *xRearSR = docHandle.FirstChildElement("car").FirstChild("rear_SR_Pacejka").FirstChild().ToText();
	TiXmlText *xRearSA = docHandle.FirstChildElement("car").FirstChild("rear_SA_Pacejka").FirstChild().ToText();
	
	// do error checking
	if (!(xCar && xTitle && xAuthor && xSprite && xSpriteFilename && xGeometry && xProperties && xWheelsprite &&
        xWheelSpriteFilename && xFrontSR && xFrontSA && xRearSR && xRearSA))
	{
		xmlDoc->Clear();
		delete xmlDoc;
      ScThrowErr("xml element that is required was not found");
	}
   
   strncpy(title, xTitle->Value(), 255);
   strncpy(author, xAuthor->Value(), 255);
   
   TiXmlNode *xConvex;
   short count;
   geometry.clear();
   for (xConvex = xGeometry->FirstChild(); xConvex != NULL; xConvex = xConvex->NextSibling("convex")) {
      geometry.push_back(std::list<Vec2D>::list());
      std::list<Vec2D> &convex = geometry.back();
      readPointsFromString(xConvex->FirstChild()->ToText()->Value(), convex, count);
   }
   
   readDataFromString(xFrontSR->Value(), frontSlipRatioPacejka, 4);
   readDataFromString(xFrontSA->Value(), frontSlipAnglePacejka, 4);
   readDataFromString(xRearSR->Value(), rearSlipRatioPacejka, 4);
   readDataFromString(xRearSA->Value(), rearSlipAnglePacejka, 4);
   
   if (xSprite->QueryIntAttribute("framewidth", &vehicleFrameWidth) != TIXML_SUCCESS)
      ScThrowErr("xml element sprite property framewidth not defined");
   if (xSprite->QueryIntAttribute("frameheight", &vehicleFrameHeight) != TIXML_SUCCESS)
      ScThrowErr("xml element sprite property frameheight not defined");
   
   if (xWheelsprite->QueryIntAttribute("framewidth", &wheelFrameWidth) != TIXML_SUCCESS)
      ScThrowErr("xml element wheelsprite property framewidth not defined");
   if (xWheelsprite->QueryIntAttribute("frameheight", &wheelFrameHeight) != TIXML_SUCCESS)
      ScThrowErr("xml element wheelsprite property frameheight not defined");
   const char *str = xWheelsprite->Attribute("layer");
   if (!str)
      ScThrowErr("xml element wheelsprite property layer not defined");
   if (strcmp(str,"below") == 0)
      wheelsAboveSprite = false;
   else if (strcmp(str,"above") == 0)
      wheelsAboveSprite = true;
   else
      ScThrowErr("xml element wheelsprite property layer not defined correctly (must be either \"above\" or \"below\"");
   
   if (xProperties->QueryDoubleAttribute("axelwidth", &axelWidth) != TIXML_SUCCESS)
      ScThrowErr("xml element property axelwidth not defined");
   if (xProperties->QueryDoubleAttribute("frontaxel", &frontAxel) != TIXML_SUCCESS)
      ScThrowErr("xml element property frontaxel not defined");
   if (xProperties->QueryDoubleAttribute("rearaxel", &rearAxel) != TIXML_SUCCESS)
      ScThrowErr("xml element property rearaxel not defined");
   if (xProperties->QueryDoubleAttribute("frontwheelinset", &frontWheelInset) != TIXML_SUCCESS)
      ScThrowErr("xml element property frontwheelinset not defined");
   if (xProperties->QueryDoubleAttribute("frontwheelradius", &frontWheelRadius) != TIXML_SUCCESS)
      ScThrowErr("xml element property frontwheelradius not defined");
   if (xProperties->QueryDoubleAttribute("rearwheelradius", &rearWheelRadius) != TIXML_SUCCESS)
      ScThrowErr("xml element property rearwheelradius not defined");
   if (xProperties->QueryDoubleAttribute("mass", &mass) != TIXML_SUCCESS)
      ScThrowErr("xml element property mass not defined");
   if (xProperties->QueryDoubleAttribute("inertia", &inertia) != TIXML_SUCCESS)
      ScThrowErr("xml element property inertia not defined");
   if (xProperties->QueryDoubleAttribute("frontwheelinertia", &frontWheelInertia) != TIXML_SUCCESS)
      ScThrowErr("xml element property frontwheelinertia not defined");
   if (xProperties->QueryDoubleAttribute("rearwheelinertia", &rearWheelInertia) != TIXML_SUCCESS)
      ScThrowErr("xml element property rearwheelinertia not defined");
   if (xProperties->QueryDoubleAttribute("drag", &drag) != TIXML_SUCCESS)
      ScThrowErr("xml element property drag not defined");
   if (xProperties->QueryDoubleAttribute("rollingresistance", &rollingResistance) != TIXML_SUCCESS)
      ScThrowErr("xml element property rollingresistance not defined");
	
   char *temp;
   
   // get the vehicle path name
   if (strncmp(xSpriteFilename->Value(), "$data", 5) == 0) {
      strncpy(vehicleSpriteFilename, xSpriteFilename->Value(), 255);
   }
   else {
      strncpy(vehicleSpriteFilename, fullFilename, 255);
      temp = (char *)strrchr(vehicleSpriteFilename, '/');
      if (!temp)
         ScThrowErr("xml vehicle sprite filename is not absolute");
      temp[1] = '\0';
      strcat(vehicleSpriteFilename, xSpriteFilename->Value());
   }
   
   // get the wheel path name
   if (strncmp(xWheelSpriteFilename->Value(), "$data", 5) == 0) {
      strncpy(wheelSpriteFilename, xWheelSpriteFilename->Value(), 255);
   }
   else {
      strncpy(wheelSpriteFilename, fullFilename, 255);
      temp = (char *)strrchr(wheelSpriteFilename, '/');
      if (!temp)
         ScThrowErr("xml wheel sprite filename is not absolute");
      temp[1] = '\0';
      strcat(wheelSpriteFilename, xWheelSpriteFilename->Value());
   }
   
   // Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
}

void VehicleObject::saveVehicle(const char *filename)
{
	// Save the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(filename);
	char str[1024];
	char vehicleSpriteLocalName[1024];
	char wheelSpriteLocalName[1024];
   char *c;
   bool vehicleSpriteLocal, wheelSpriteLocal;
   
   TiXmlDeclaration decl("1.0", "UTF-8", "yes");
	TiXmlElement xCar("car");
	TiXmlElement xTitle("title");
	TiXmlElement xAuthor("author");
	TiXmlElement xSprite("sprite");
	TiXmlElement xGeometry("geometry");
	TiXmlElement xProperties("properties");
	TiXmlElement xWheelsprite("wheelsprite");
	TiXmlElement xFrontSR("front_SR_Pacejka");
	TiXmlElement xFrontSA("front_SA_Pacejka");
	TiXmlElement xRearSR("rear_SR_Pacejka");
	TiXmlElement xRearSA("rear_SA_Pacejka");
	
	xTitle.InsertEndChild(TiXmlText(title));
   
	xAuthor.InsertEndChild(TiXmlText(author));
   
   vehicleSpriteLocal = strncmp(vehicleSpriteFilename, "$data", 5) != 0;
   wheelSpriteLocal = strncmp(wheelSpriteFilename, "$data", 5) != 0;
   
   if (vehicleSpriteLocal)
      strcpy(vehicleSpriteLocalName, strrchr(vehicleSpriteFilename, '/')+1);
   else
      strcpy(vehicleSpriteLocalName, vehicleSpriteFilename);
   
   if (wheelSpriteLocal)
      strcpy(wheelSpriteLocalName, strrchr(wheelSpriteFilename, '/')+1);
   else
      strcpy(wheelSpriteLocalName, wheelSpriteFilename);
   
   xSprite.SetDoubleAttribute("framewidth", vehicleFrameWidth);
	xSprite.SetDoubleAttribute("frameheight", vehicleFrameHeight);
   xSprite.InsertEndChild(TiXmlText(vehicleSpriteLocalName));
   
   for (Geometry::iterator it = geometry.begin(); it != geometry.end(); it++) {
      TiXmlElement *xConvex = new TiXmlElement("convex");
      c = str;
      for (std::list<Vec2D>::iterator ot = it->begin(); ot != it->end(); ot++) {
         sprintf(c, "%f,%f; ", ot->x, ot->y);
         c += strlen(c);
      }
      xConvex->InsertEndChild(TiXmlText(str));
      xGeometry.InsertEndChild(*xConvex);
      delete xConvex;   // do this?
   }
   
   xProperties.SetDoubleAttribute("axelwidth", axelWidth);
   xProperties.SetDoubleAttribute("frontaxel", frontAxel);
   xProperties.SetDoubleAttribute("rearaxel", rearAxel);
   xProperties.SetDoubleAttribute("frontwheelinset", frontWheelInset);
   xProperties.SetDoubleAttribute("frontwheelradius", frontWheelRadius);
   xProperties.SetDoubleAttribute("rearwheelradius", rearWheelRadius);
   xProperties.SetDoubleAttribute("mass", mass);
   xProperties.SetDoubleAttribute("inertia", inertia);
   xProperties.SetDoubleAttribute("frontwheelinertia", frontWheelInertia);
   xProperties.SetDoubleAttribute("rearwheelinertia", rearWheelInertia);
   xProperties.SetDoubleAttribute("drag", drag);
   xProperties.SetDoubleAttribute("rollingresistance", rollingResistance);
   
	xWheelsprite.SetDoubleAttribute("framewidth", wheelFrameWidth);
	xWheelsprite.SetDoubleAttribute("frameheight", wheelFrameHeight);
	xWheelsprite.SetAttribute("layer", wheelsAboveSprite ? "above" : "below");
   xWheelsprite.InsertEndChild(TiXmlText(wheelSpriteLocalName));
   
   sprintf(str, "%f, %f, %f, %f", frontSlipRatioPacejka[0], frontSlipRatioPacejka[1], frontSlipRatioPacejka[2], frontSlipRatioPacejka[3]);
   xFrontSR.InsertEndChild(TiXmlText(str));
   sprintf(str, "%f, %f, %f, %f", frontSlipAnglePacejka[0], frontSlipAnglePacejka[1], frontSlipAnglePacejka[2], frontSlipAnglePacejka[3]);
   xFrontSA.InsertEndChild(TiXmlText(str));
   sprintf(str, "%f, %f, %f, %f", rearSlipRatioPacejka[0], rearSlipRatioPacejka[1], rearSlipRatioPacejka[2], rearSlipRatioPacejka[3]);
   xRearSR.InsertEndChild(TiXmlText(str));
   sprintf(str, "%f, %f, %f, %f", rearSlipAnglePacejka[0], rearSlipAnglePacejka[1], rearSlipAnglePacejka[2], rearSlipAnglePacejka[3]);
   xRearSA.InsertEndChild(TiXmlText(str));
   
   xmlDoc->InsertEndChild(decl);
   xCar.InsertEndChild(xTitle);
   xCar.InsertEndChild(xAuthor);
   xCar.InsertEndChild(xSprite);
   xCar.InsertEndChild(xGeometry);
   xCar.InsertEndChild(xProperties);
   xCar.InsertEndChild(xWheelsprite);
   xCar.InsertEndChild(xFrontSR);
   xCar.InsertEndChild(xFrontSA);
   xCar.InsertEndChild(xRearSR);
   xCar.InsertEndChild(xRearSA);
   xmlDoc->InsertEndChild(xCar);
   
	xmlDoc->SaveFile();
   
	// Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
   
   // now copy the sprite files to the proper directories...
   char cmd[1024];
   if (vehicleSpriteLocal || wheelSpriteLocal) {
      strcpy(str, filename);
      c = strrchr(str, '/');
      c[0] = '\0';
   }
   if (vehicleSpriteLocal) {
      if (strcmp(vehicleSpriteFilename, str) != 0) {
      #ifdef WIN32
         fl_alert("Please copy %s to %s. Sorry on windows this must be done manually for now.", vehicleSpriteLocalName, str);
      #else
         sprintf(cmd, "cp %s %s", vehicleSpriteFilename, str);
         system(cmd);
      #endif
      }
   }
   if (wheelSpriteLocal) {
      if (strcmp(wheelSpriteFilename, str) != 0) {
      #ifdef WIN32
         fl_alert("Please copy %s to %s. Sorry on windows this must be done manually for now.", wheelSpriteLocalName, str);
      #else
         sprintf(cmd, "cp %s %s", wheelSpriteFilename, str);
         system(cmd);
      #endif
      }
   }
}

void VehicleObject::loadSpriteFromDataDir(const char *name)
{
   char str[1024] = "cars/";
   char path[1024];
   
   strcat(str, name);
   strcat(str, "/");
   strcat(str, name);
   strcat(str, ".xml");
   
   if (!findDataFile(str, gSearchDirs, path))
      ScThrowErr("file not found");
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(path);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandle(xmlDoc);
	
	TiXmlElement *xCar = docHandle.FirstChildElement("car").Element();
	TiXmlElement *xSprite = docHandle.FirstChildElement("car").FirstChildElement("sprite").Element();
	TiXmlText *xSpriteFilename = xSprite->FirstChild()->ToText();
	TiXmlElement *xGeometry = docHandle.FirstChildElement("car").FirstChildElement("geometry").Element();
	TiXmlElement *xProperties = docHandle.FirstChildElement("car").FirstChildElement("properties").Element();
	
	// do error checking
	if (!(xCar && xSprite && xSpriteFilename && xGeometry && xProperties))
	{
		xmlDoc->Clear();
		delete xmlDoc;
      ScThrowErr("xml element that is required was not found");
	}
   
   TiXmlNode *xConvex;
   short count;
   geometry.clear();
   for (xConvex = xGeometry->FirstChild(); xConvex != NULL; xConvex = xConvex->NextSibling("convex")) {
      geometry.push_back(std::list<Vec2D>::list());
      std::list<Vec2D> &convex = geometry.back();
      readPointsFromString(xConvex->FirstChild()->ToText()->Value(), convex, count);
   }
   
   if (xSprite->QueryIntAttribute("framewidth", &vehicleFrameWidth) != TIXML_SUCCESS)
      ScThrowErr("xml element sprite property framewidth not defined");
   if (xSprite->QueryIntAttribute("frameheight", &vehicleFrameHeight) != TIXML_SUCCESS)
      ScThrowErr("xml element sprite property frameheight not defined");
   
   if (xProperties->QueryDoubleAttribute("axelwidth", &axelWidth) != TIXML_SUCCESS)
      ScThrowErr("xml element property axelwidth not defined");
   if (xProperties->QueryDoubleAttribute("frontaxel", &frontAxel) != TIXML_SUCCESS)
      ScThrowErr("xml element property frontaxel not defined");
   if (xProperties->QueryDoubleAttribute("rearaxel", &rearAxel) != TIXML_SUCCESS)
      ScThrowErr("xml element property rearaxel not defined");
   
   // get the vehicle path name
   if (strncmp(xSpriteFilename->Value(), "$data", 5) == 0) {
      strncpy(vehicleSpriteFilename, xSpriteFilename->Value(), 255);
   }
   else {
      strcpy(vehicleSpriteFilename, "$data/cars/");
      strcat(vehicleSpriteFilename, name);
      strcat(vehicleSpriteFilename, "/");
      strcat(vehicleSpriteFilename, xSpriteFilename->Value());
   }
   
   // Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
}

void VehicleObject::loadWheelSpriteFromDataDir(const char *name)
{
   char str[1024] = "cars/";
   char path[1024];
   
   strcat(str, name);
   strcat(str, "/");
   strcat(str, name);
   strcat(str, ".xml");
   
   if (!findDataFile(str, gSearchDirs, path))
      ScThrowErr("file not found");
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(path);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	TiXmlHandle docHandle(xmlDoc);
	
	TiXmlElement *xCar = docHandle.FirstChildElement("car").Element();
	TiXmlElement *xWheelsprite = docHandle.FirstChildElement("car").FirstChildElement("wheelsprite").Element();
	TiXmlText *xWheelSpriteFilename = xWheelsprite->FirstChild()->ToText();
	TiXmlElement *xProperties = docHandle.FirstChildElement("car").FirstChildElement("properties").Element();
	
	// do error checking
	if (!(xCar && xWheelsprite && xWheelSpriteFilename && xProperties))
	{
		xmlDoc->Clear();
		delete xmlDoc;
      ScThrowErr("xml element that is required was not found");
	}
      
   if (xWheelsprite->QueryIntAttribute("framewidth", &wheelFrameWidth) != TIXML_SUCCESS)
      ScThrowErr("xml element wheelsprite property framewidth not defined");
   if (xWheelsprite->QueryIntAttribute("frameheight", &wheelFrameHeight) != TIXML_SUCCESS)
      ScThrowErr("xml element wheelsprite property frameheight not defined");
   const char *layerStr = xWheelsprite->Attribute("layer");
   if (!layerStr)
      ScThrowErr("xml element wheelsprite property layer not defined");
   if (strcmp(layerStr,"below") == 0)
      wheelsAboveSprite = false;
   else if (strcmp(layerStr,"above") == 0)
      wheelsAboveSprite = true;
   else
      ScThrowErr("xml element wheelsprite property layer not defined correctly (must be either \"above\" or \"below\"");
   
   if (xProperties->QueryDoubleAttribute("frontwheelinset", &frontWheelInset) != TIXML_SUCCESS)
      ScThrowErr("xml element property frontwheelinset not defined");
   if (xProperties->QueryDoubleAttribute("frontwheelradius", &frontWheelRadius) != TIXML_SUCCESS)
      ScThrowErr("xml element property frontwheelradius not defined");
   if (xProperties->QueryDoubleAttribute("rearwheelradius", &rearWheelRadius) != TIXML_SUCCESS)
      ScThrowErr("xml element property rearwheelradius not defined");
	   
   // get the vehicle path name
   if (strncmp(xWheelSpriteFilename->Value(), "$data", 5) == 0) {
      strncpy(wheelSpriteFilename, xWheelSpriteFilename->Value(), 255);
   }
   else {
      strcpy(wheelSpriteFilename, "$data/cars/");
      strcat(wheelSpriteFilename, name);
      strcat(wheelSpriteFilename, "/");
      strcat(wheelSpriteFilename, xWheelSpriteFilename->Value());
   }
   
   // Clear and delete our allocated document and return success ;)
	xmlDoc->Clear();
	delete xmlDoc;
}

// installs map into user data directory
void VehicleObject::installVehicle(const char *name)
{
#ifdef WIN32
   fl_alert("Unfortunately this is unsupported on Windows at this time.\nYou may still install vehicles manually by saving them to a folder and placing it inside the data\\cars directory and editing the carslist.xml file.");
   return;
#else
   char fullname[1024];
   char carname[256];
   const char kVehicleListName[64] = "/cars/carslist.xml";
   char *temp;
   
   // first we uninstall the named car
   uninstallVehicle(name);
   
   // chop off any .xml suffix
   temp = strrchr(name, '.');
   if (temp) {
      if (strcmp(temp, ".xml") == 0)
         temp[0] = '\0';
   }
   // remove path
   temp = strrchr(name, '/');
   if (temp) {
      strncpy(carname, temp+1, 255);
   }
   
   char userDataDir[1024];
   sprintf(userDataDir, "%s/.toycars", getenv("HOME"));
   strncpy(fullname, userDataDir, 1023 - strlen(kVehicleListName));
   strcat(fullname, kVehicleListName);
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullname);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	// Add name to car list
	TiXmlHandle *docHandle = new TiXmlHandle(xmlDoc);
	TiXmlElement *xCarslist = docHandle->FirstChildElement("carslist").Element();
	if (xCarslist == NULL)
		ScThrowErr("xml file could not find carslist element");
   TiXmlElement car("car");
   TiXmlText nameText(carname);
   car.InsertEndChild(nameText);
   xCarslist->InsertEndChild(car);
	
	// Save the xml file
	if (!xmlDoc->SaveFile())
	{
		delete xmlDoc;
		ScThrowErr("error saving xml file");
	}
	
	//delete xCarslist;
	delete docHandle;
	delete xmlDoc;
	
   strncpy(fullname, userDataDir, 1023 - 8);
   strcat(fullname, "/cars/");
   strcat(fullname, carname);

   // make sure the cars directory exists. nb: what whould be done on win32 here???
   char str[1024];
   sprintf(str, "mkdir -p %s", fullname);
   system(str);

   strcat(fullname, "/");
   strcat(fullname, carname);
   strcat(fullname, ".xml");
   
   // save it to the data directory
   saveVehicle(fullname);
#endif
}

// uninstall car from user data directory.
// nb: doesn't remove car data itself, only remove entry from carslist xml file.
void VehicleObject::uninstallVehicle(const char *name)
{
#ifdef WIN32
   fl_alert("Unfortunately this is unsupported on Windows at this time.\nYou may still uninstall vehicles manually by removing them from the data\\cars directory and editing the carslist.xml file.");
   return;
#else
   char fullname[1024];
   char carname[256];
   const char kCarslistName[64] = "/cars/carslist.xml";
   char *temp;
   
   // chop off any .xml suffix
   temp = strrchr(name, '.');
   if (temp) {
      if (strcmp(temp, ".xml") == 0)
         temp[0] = '\0';
   }
   // remove path
   temp = strrchr(name, '/');
   if (temp) {
      strncpy(carname, temp+1, 255);
   }
   
   char userDataDir[1024];
   sprintf(userDataDir, "%s/.toycars", getenv("HOME"));
   strncpy(fullname, userDataDir, 1023 - strlen(kCarslistName));
   strcat(fullname, kCarslistName);
   
	// Load the xml file
	TiXmlDocument *xmlDoc = new TiXmlDocument(fullname);
	if (!xmlDoc->LoadFile())
	{
		delete xmlDoc;
		ScThrowErr("xml file not found");
	}
	
	// Remove all occurrences of the car the from car list
	TiXmlHandle *docHandle = new TiXmlHandle(xmlDoc);
	TiXmlElement *xCarslist = docHandle->FirstChildElement("carslist").Element();
	if (xCarslist == NULL)
		ScThrowErr("xml file could not find carlist element");
   TiXmlElement *car = xCarslist->FirstChildElement("car");
   TiXmlElement *prevCar;
	while (car != NULL)
	{
      const char *str = car->FirstChild()->ToText()->Value();
		if (str == NULL)
			ScThrowErr("xml car element does not contain text");
		prevCar = car;
		car = car->NextSiblingElement("car");
      if (strcmp(str, carname) == 0) {
         xCarslist->RemoveChild(prevCar);
      }
	}
   
   // Save the xml file
   if (!xmlDoc->SaveFile()) {
      delete xmlDoc;
      ScThrowErr("error saving xml file");
   }
	
	//delete xCarslist;
	delete docHandle;
	delete xmlDoc;
   
   /*
   // this could be dangerous, so I'll comment it out.
   // remove the car directory. nb: what whould be done on win32 here???
   strncpy(fullname, userDataDir, 1023 - 8);
   strcat(fullname, "/cars/");
   strcat(fullname, carname);

   char str[1024];
   sprintf(str, "rm -rf %s", fullname);
   system(str);
   */
#endif
}